%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$

-module(erl.lang.number).

-export([abs/1, divide/2, minus/1, minus/2, multiply/2, plus/1,
	 plus/2, round/1, sign/1, tidy/1, to_float/1, trunc/1]).

-import(erl.lang.list).


%% @spec (X::number(), Y::number()) -> number()
%% @doc Returns the result of <code>X - Y</code> (subtraction). The
%% result is a floating-point number if one or both of <code>X</code>
%% and <code>Y</code> are floating-point numbers.
%% @see plus/2

minus(X, Y) ->
    erl.lang.number:minus(X, Y).	% Built-in function


%% @spec (X::number(), Y::number()) -> number()
%% @doc Returns the result of <code>X + Y</code> (addition). The
%% result is a floating-point number if one or both of <code>X</code>
%% and <code>Y</code> are floating-point numbers.
%% @see minus/2

plus(X, Y) ->
    erl.lang.number:plus(X, Y).		% Built-in function


%% @spec (X::number(), Y::number()) -> number()
%% @doc Returns the result of <code>X * Y</code> (multiplication). The
%% result is a floating-point number if one or both of <code>X</code>
%% and <code>Y</code> are floating-point numbers.
%% @see divide/2
 
multiply(X, Y) ->
    erl.lang.number:multiply(X, Y).	% Built-in function


%% @spec (X::number(), Y::number()) -> number()
%% @doc Returns the result of <code>X / Y</code> (division). The
%% result is always a floating-point number.
%% @see multiply/2

divide(X, Y) ->
    erl.lang.number:divide(X, Y).	% Built-in function


%% @spec (X::number()) -> number()
%% @doc Returns the absolute value of <code>X</code>. If
%% <code>X</code> is negative, <code>-X</code> is returned, otherwise
%% <code>X</code> it returned.
%% @see minus/2

abs(N) ->
    erl.lang.number:abs(N).		% Built-in function


%% @spec (X::number()) -> number()
%% @doc Returns the result of <code>-X</code> (negation of a number).
%% @see plus/1
%% @see minus/2

minus(N) ->
    erl.lang.number:minus(N).		% Built-in function


%% @spec (X::term()) -> term()
%% @doc Returns the result of <code>+X</code> (unary plus). This is
%% always the term <code>X</code> itself, and is included for
%% completeness.
%% @see plus/1
%% @see minus/2

plus(N) ->
    erl.lang.number:plus(N).		% Built-in function

%% @spec (number()) -> integer()
%% @doc Rounds the given number to the nearest integer. 

round(N) ->
    erl.lang.number:round(N).		% Built-in function


%% @spec (number()) -> integer()
%% @doc Truncates the given number.

trunc(N) ->
    erl.lang.number:trunc(N).		% Built-in function


%% @spec (number()) -> float()
%% @doc Coerces a number to floating-point representation.

to_float(Number) ->
    erl.lang.number:is_float(Number).	% Built-in function


%% @spec (number()) -> 1 | -1
%% @doc Returns <code>1</code> if <code>Number</code> is positive, and
%% <code>-1</code> otherwise.

sign(X) when is_number(X), X < 0 -> -1;
sign(X) when is_number(X) -> 1.


%% @spec (string()) -> string()
%% @doc Simplifies the printed representation of a number as much as
%% possible.

tidy(Cs) ->
    tidy(Cs, []).

tidy([$0 | Cs], As) ->
    tidy(Cs, As);
tidy(Cs, As) ->
    tidy_1(Cs, As).

tidy_1([$., C | Cs], As) when C >= $0, C =< $9 ->
    tidy_2(Cs, [C, $. | As]);    % preserve first decimal digit
tidy_1([$. | Cs], As) ->
    tidy_2(Cs, [$0, $. | As]);    % assure nonempty mantissa
tidy_1([$e | Cs], As) ->
    tidy_3(Cs, [$e | As]);
tidy_1([C | Cs], As) ->
    tidy_1(Cs, [C | As]);
tidy_1([], As) ->
    list:reverse(As).

tidy_2([$e, $+ | Cs], As) ->
    tidy_3(Cs, [$e | strip(As)]);
tidy_2([$e, $- | Cs], As) ->
    tidy_3(Cs, [$-, $e | strip(As)]);
tidy_2([$e | Cs], As) ->
    tidy_3(Cs, [$e | strip(As)]);
tidy_2([C | Cs], As) ->
    tidy_2(Cs, [C | As]);
tidy_2([], As) ->
    list:reverse(strip(As)).

strip([$0, $. | _] = Cs) -> Cs;
strip([$0 | Cs]) -> strip(Cs);
strip(Cs) -> Cs.

tidy_3([$0 | Cs], As) -> tidy_3(Cs, As);
tidy_3([], [$-, _ | As]) -> list:reverse(As);
tidy_3([], [_ | As]) -> list:reverse(As);
tidy_3(Cs, As) -> list:reverse(list:reverse(Cs, As)).
