%%-------------------------------------------------------------------
%% File    : cerl_bin_.erl
%% Author  : Per Gustafsson <pergu@it.uu.se>
%% Description : 
%%
%% Created : 28 May 2003 by Per Gustafsson <pergu@it.uu.se>
%%-------------------------------------------------------------------

-module(cerl_binary_pattern_match).
-author('pergu@it.uu.se').

-export([add_offset_to_bin/1]).
-export([read_seg_size/1, read_seg_offset/1, read_seg_type/1,
	 read_seg_tag/1, match_tag/1, match_val/1, match_group_tag/1,
	 match_group_vals/1,
	 binary_match_clause_tree/1, binary_match_max_tag/1,
	 clause_tree_instr/1, clause_tree_success/1, clause_tree_fail/1,
	 bin_guard_matches/1, bin_guard_label/1,
	 size_const/1, size_all/1, size_vars/1,size_def_vars/1,
	 instr_type/1, label_name/1, goto_label/1]).
%%-export([pp/1, full_size/1, a_height/1, max_height/1, read_seg/4]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Record definitions and access functions 
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% @type instr()= size() | read_seg() | match() | label() | goto() | bin_guard() .
%% All these data types can be used as instructions 

%% @type read_seg(). an abstract data type containing type,
%% offset, size and tag data which describes a read segment
-record(read_seg, {type, offset, size, tag}).

%% @spec read_seg_size(read_seg()) -> integer() | var()
%% @doc Returns the size field from a read_seg
read_seg_size(#read_seg{size=X}) -> X.

%% @spec read_seg_offset(read_seg()) -> size()
%% @doc Returns the offset field from a read_seg
read_seg_offset(#read_seg{offset=X}) -> X.

%% @spec read_seg_type(read_seg()) -> {atom(), atom(), atom()}
%% @doc Returns the type field from a read_seg
read_seg_type(#read_seg{type=X}) -> X.

%% @spec read_seg_tag(read_seg()) -> integer()
%% @doc Returns the tag field from a read_seg
read_seg_tag(#read_seg{tag=X}) -> X.

%% @spec read_seg({atom(), atom(), atom()}, size(), integer() | var(), integer()) -> read_seg()
%% @doc Creates a read_seg from a type an offset a size and a tag
read_seg(Type, Offset, Size, Tag)->
  #read_seg{type=Type, offset=Offset, size=Size, tag=Tag}.

%% @type b_clause(). Represents an original binary clause consisting of
%% instructions, matches to be performed at the guard, a guard, the body
%% of the clause and a pointer to the next clause this is only used internally

-record(b_clause, {segments, matches=[], guard, body, next_clause}).
b_clause_segments(#b_clause{segments=X}) -> X.


%% @type match(). Reprsents a matching between a result from a read_seg
%% which is represented as a tag and a term which is called val
-record(match, {val, tag}).
%% @spec match_tag(match()) -> integer()
%% @doc Returns the tag field from a match test
match_tag(#match{tag=X})-> X.
%% @spec match_val(match()) -> term()
%% @doc Returns the val field from a match test
match_val(#match{val=X})-> X.

-record(match_group, {tag, vals}).
match_group_tag(#match_group{tag=X})->X.
match_group_vals(#match_group{vals=X}) ->X.

%% @type binary_match(). Represents an entire binary matching, The structure 
%% of the matching is described in the clause tree and the max tag shows
%% how many reads might be needed
-record(binary_match, {clause_tree, max_tag}).
%% @spec binary_match_clause_tree(binary_match()) -> clause_tree()
%% @doc Returns the clause_tree field from a binary match 
binary_match_clause_tree(#binary_match{clause_tree=X}) -> X.

%% @spec binary_match_max_tag(binary_match()) -> int()
%% @doc Returns themax_tag field from a binary match 
binary_match_max_tag(#binary_match{max_tag=X}) -> X.

%% @type clause_tree(). A clause tree is a tree of instructions
%% each node contains an instruction and a pointer to the success tree
%% and one to the fail tree
-record(clause_tree, {instr, success=[], fail=[]}).
%% @spec clause_tree_instr(binary_match()) -> instr()
%% @doc Returns the instr field from a clause tree 
clause_tree_instr(#clause_tree{instr=X}) -> X.

%% @spec clause_tree_success(clause_tree()) -> clause_tree()
%% @doc Returns the succesful clause tree rooted in this node 
clause_tree_success(#clause_tree{success=X}) -> X.

%% @spec clause_tree_fail(clause_tree()) -> clause_tree()
%% @doc Returns the failure clause tree rooted in this node 
clause_tree_fail(#clause_tree{fail=X}) -> X.

%% @type bin_guard(). Represents a guard for a binary clause
%% containing a set of necessery matches and an outside name
-record(bin_guard, {matches, label}).
%% @spec bin_guard_matches(bin_guard()) -> [match()]
%% @doc Returns the set of matches that need to be performed
%% when this guard is reached 
bin_guard_matches(#bin_guard{matches=X}) -> X.

%% @spec bin_guard_label(bin_guard()) -> [ref()]
%% @doc Returns the outside label name of the guard
bin_guard_label(#bin_guard{label=X}) -> X.

%% @type size(). Represents a size expression containg a
%% constant part, an "operator" part and two different
%% variable parts one for locally defined variables and
%% one for other variables
-record(size, {vars=[],  def_vars=[],const=cerl:c_int(0),  all=cerl:c_atom(false)}).
%% @spec size_const(size()) -> cerl:c_int()
%% @doc Returns the constant part of a size expression
size_const(#size{const=X})-> X.

%% @spec size_all(size()) -> cerl:c_atom(false) | cerl:c_atom(all)
%% @doc Returns cerl:c_atom(all) if the demand on the size is greater than
%% returns cerl:c_atom(false) in other cases
size_all(#size{all=X})-> X.

%% @spec size_vars(size()) -> [cerl:c_var()]
%% @doc Returns a list of the non-locally defined 
%% variables of a size expression
size_vars(#size{vars=X})-> X.

%% @spec size_def_vars(size()) -> [cerl:c_var()]
%% @doc Returns a list of the locally defined 
%% variables of a size expression
size_def_vars(#size{def_vars=X})-> X.

%% @spec size_all_vars(size()) -> [cerl:c_var()]
%% @doc Returns a list of all 
%% variables of a size expression
size_all_vars(#size{vars=X,def_vars=Y})-> X++Y.

%% @type label(). The label data type represents a binary match local label
-record(label, {name}).
%% @spec label_name(label()) -> integer()
%% @doc returns the name of a label
label_name(#label{name=X}) -> X.

%% @type goto(). The goto data type represents a binary match local goto 
%% instruction
-record(goto, {label}).
%% @spec goto_label(goto()) -> integer()
%% @doc returns the name of a label which the goto points to
goto_label(#goto{label=X}) -> X.


-record(hash_cons, {tree=gb_trees:empty(), number=0}).

new_node(Key,Hash=#hash_cons{tree=Tree, number=Number}) ->
  case gb_trees:lookup(Key, Tree) of
    none ->
      LabelName=cerl:c_int(Number),
      {new, LabelName, Hash#hash_cons{tree=gb_trees:insert(Key, LabelName, Tree), number=Number+1}};
    {value, LabelName} ->
      {already_exists, LabelName}
  end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%%End of record definitions
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%========================================================================
%% @spec add_offset_to_bin(Binclauses::[clauses()]) -> binary_match()
%% 
%% binary_match() = cerl_binary_pattern_match:binary_match()
%%
%% @doc Turns a list of binclauses into a
%% binary match consisting of a clause tree and a the highest tag value
%% @end
%%========================================================================

add_offset_to_bin(BinClauses) ->
  {BClause0, MaxTag}=iterate_binclauses(BinClauses),
  {ClauseTree, _Hash}=adapt(BClause0, #hash_cons{}),
  #binary_match{clause_tree=ClauseTree, max_tag=MaxTag}.

%%==========================================================================
%%
%% iterate_binclauses(Clauses) Frees segments from their context and creates
%% tags to show which segments that are similar, returns a b_clause and 
%% the first unused tag
%%
%%==========================================================================

iterate_binclauses(Clauses) ->
  iterate_binclauses(Clauses, 0, gb_trees:empty()).

iterate_binclauses([Clause|Rest], Tag, TagMap) ->
  Guard = cerl:clause_guard(Clause),
  Body = cerl:clause_body(Clause),
  [Bin] = cerl:clause_pats(Clause),
  BinSegs = cerl:binary_segments(Bin),
  {Matches, NewBinSegs, NewTag, NewTagMap} = 
    iterate_binsegs(BinSegs, Tag, TagMap), 
  {Next, FinalTag} = iterate_binclauses(Rest, NewTag, NewTagMap),
  {#b_clause{segments=NewBinSegs, matches=Matches, guard=Guard,  
	    body=Body, next_clause = Next}, FinalTag};
iterate_binclauses([], Tag, _TagMap) ->
  {[], Tag}.

iterate_binsegs(BinSegs, Tag, TagMap) ->
  iterate_binsegs(BinSegs,  Tag, [], #size{}, [], [], TagMap).
  

iterate_binsegs([Seg|Rest], Tag, AccSegs, AccSize, AccMatches, AccDefs, TagMap) ->
  case annotate_seg(Seg, Tag, AccSize, AccDefs, TagMap) of
    {var, {{NewTag, NewTagMap}, ReadNode, Match}, NewAccSize} ->
      iterate_binsegs(Rest, NewTag, [ReadNode|AccSegs], NewAccSize, [Match|AccMatches], [Match|AccDefs], NewTagMap);
    {const, {{NewTag, NewTagMap}, ReadNode, MatchNode}, NewAccSize} ->
      iterate_binsegs(Rest, NewTag, [MatchNode,ReadNode|AccSegs], NewAccSize, AccMatches, AccDefs, NewTagMap);
    {size, {SizeTest,{var, {{NewTag, NewTagMap}, ReadNode, Match}, NewAccSize}}} ->
      iterate_binsegs(Rest, NewTag, [SizeTest,ReadNode|AccSegs], NewAccSize, [Match|AccMatches], [Match], NewTagMap);
    {size, {SizeTest, {const, {{NewTag, NewTagMap}, ReadNode, MatchNode}, NewAccSize}}}->
      iterate_binsegs(Rest, NewTag, [SizeTest,MatchNode,ReadNode|AccSegs], NewAccSize, AccMatches, [], NewTagMap)
  end;

iterate_binsegs([], Tag, AccSegs, AccSize, AccMatches, _AccDefs, TagMap) ->
  {AccMatches, [AccSize|lists:reverse(AccSegs)], Tag, TagMap}.

annotate_seg(Segment, Tag, Offset, Defs, TagMap) ->
  Type = seg_type(Segment),
  Size = cerl:bitstr_size(Segment),
  Unit = cerl:bitstr_unit(Segment),
  Key=create_key(Type, Size, Unit, Offset),
  {ThisTag, UpdatedTags}=find_tag(Key, Tag, TagMap),
  Val = cerl:bitstr_val(Segment),
  Match = #match{val=Val, tag=ThisTag},
  case size_var_def(Size, Defs) of 
    {true, SizeTag} ->
      NewSize = combine(SizeTag, Unit),
      NextOffset = calc_offset(NewSize,Offset),
      AllOffset = Offset#size{all=cerl:c_atom(all)},
      Read = read_seg(Type, Offset, NewSize, ThisTag),
      Next = case cerl:is_c_var(Val) of
	       true ->
		 {var, {UpdatedTags, Read, Match}, NextOffset};
	       false ->
		 {const, {UpdatedTags, Read, Match}, NextOffset}
	     end,
      {size, {AllOffset, Next}};
    false ->
      NewSize = combine(Size, Unit),
      NextOffset = calc_offset(NewSize,Offset),
      Read = read_seg(Type, Offset, NewSize, ThisTag),
      case cerl:is_c_var(Val) of
	true ->
	  {var, {UpdatedTags, Read, Match}, NextOffset};
	false ->
	  {const, {UpdatedTags, Read, Match}, NextOffset}
      end
  end.

create_key({Sort, Flags}, Size, Unit, Offset) ->
  {cerl:concrete(Sort),
   [cerl:concrete(X)||X<-Flags],
   sizekey(Size),
   cerl:concrete(Unit),
   offsetkey(Offset)}.

sizekey(Size) ->
  case cerl:is_c_var(Size) of
    true ->
      cerl:var_name(Size);
    false ->
      cerl:concrete(Size)
  end.

offsetkey(Offset) ->
  {cerl:concrete(size_const(Offset)),
   [{cerl:var_name(X), cerl:concrete(Y)}||{X,Y} <- size_def_vars(Offset)],
   [{cerl:var_name(X), cerl:concrete(Y)}||{X,Y} <- size_vars(Offset)],
   cerl:concrete(size_all(Offset))}.
  
find_tag(Key, PresTag, TagMap) ->
  case gb_trees:lookup(Key, TagMap) of
    {value, Tag} ->
      {Tag, {PresTag, TagMap}};
    none ->
      {PresTag, {PresTag+1, gb_trees:insert(Key, PresTag, TagMap)}}
  end.

size_var_def(Size, [Def|Rest]) ->
  case match_val(Def) of
    Size ->
      Tag=match_tag(Def),
      {true,{tag, Tag}};
    _ ->
      size_var_def(Size, Rest)
  end;

size_var_def(_Size, []) ->
  false.

%%========================================================================
%%
%% adapt(BClause, Hash) creates the clause tree dag by synthesizing
%% a traversal order and pruning all irrelevant segments at each iteration
%%
%%=========================================================================

adapt([], Hash)->
  {[], Hash};

adapt(BClause, Hash) ->
  BClause0=BClause,
  case new_node(BClause0, Hash) of
    {new, NewLabelName, Hash1} ->
      Lbl=#label{name=NewLabelName},
      {Succ, Hash2}  = adapt0(BClause0, Hash1),
      {#clause_tree{instr=Lbl, success=Succ}, Hash2};
    {already_exists, LabelName} ->
      Goto=#goto{label=LabelName},
      {#clause_tree{instr=Goto}, Hash}
  end.
      

adapt0(#b_clause{segments=[], matches=Matches, guard=Guard, body=Body, next_clause=NextClause}, Hash) ->
  {Fail, Hash1} = adapt(NextClause, Hash),
  {#clause_tree{instr=#bin_guard{matches=Matches,label=Guard}, success=Body, fail=Fail}, Hash1};

adapt0(BClause, Hash) ->
  BinSeg = choose_binseg(BClause),
  case BinSeg of
    #match{} ->
      MatchSet=get_all_similar_match(BinSeg,BClause),
      case gb_sets:size(MatchSet) of
	1 ->
	  NewSuccClause=remove_succ_match(BinSeg, BClause),
	  {Succ, Hash1} = adapt(NewSuccClause, Hash),
	   NewFailClause = remove_fail_match(BinSeg, BClause),
	  {Fail, Hash2} = adapt(NewFailClause, Hash1),
	  {#clause_tree{instr=BinSeg, success=Succ, fail=Fail}, Hash2};
	_ ->
	  {SuccList,NewInstr,Hash1}=create_success_list(MatchSet, Hash, BClause),
	  NewFailClause = remove_all_fail_match(MatchSet, BClause),
	  {Fail, Hash2} = adapt(NewFailClause, Hash1),
	  {#clause_tree{instr=NewInstr, success=SuccList, fail=Fail}, Hash2}
      end;
    #read_seg{} ->
      NewClause = remove_seg(BinSeg, BClause),
      {Succ, Hash1} = adapt(NewClause, Hash),
      {#clause_tree{instr=BinSeg, success=Succ}, Hash1};
    #size{} -> 
      {Succ, Hash1} = adapt(prune_compatible_size(BClause, BinSeg), Hash),
      {Fail, Hash2} = adapt(prune_incompatible_size(BClause, BinSeg), Hash1),
      {#clause_tree{instr=BinSeg, success=Succ, fail=Fail}, Hash2} 
  end.


get_all_similar_match(Match, BClause) ->
  List=get_all_similar_match(Match, BClause, []),
  gb_sets:from_list(List).

get_all_similar_match(Match, #b_clause{segments=BinSegs, next_clause=Next}, Acc) ->
  NewAcc=get_match(Match, BinSegs, Acc),
  get_all_similar_match(Match, Next, NewAcc);
get_all_similar_match(_Match, [], Acc) ->
  Acc.

get_match(Match1=#match{tag=Tag}, [Match2=#match{tag=Tag}|Rest], Acc) ->
  case have_same_match(Match2, Acc) of
    true ->
      get_match(Match1, Rest, Acc);
    false ->
      get_match(Match1, Rest, [Match2|Acc])
  end;
get_match(Match, [_|Rest], Acc) ->
  get_match(Match, Rest, Acc);
get_match(_Match, [], Acc) ->
  Acc.

have_same_match(Match1, [Match2|Rest]) ->
  Val1 = cerl:concrete(match_val(Match1)),
  Val2 = cerl:concrete(match_val(Match2)),
  case Val1 == Val2 of
    true ->
      true;
    false ->
      have_same_match(Match1, Rest)
  end;

have_same_match(_Match1, []) ->
  false.
    

create_success_list(MatchSet, Hash, BClause) ->
  MatchList=gb_sets:to_list(MatchSet),
  Tag=match_tag(hd(MatchList)),
  make_successor_list(MatchList, Hash, BClause, #match_group{tag=Tag, vals=[]}, 
		      gb_trees:empty()).

make_successor_list([Match=#match{val=Val}|Rest], Hash, BClause, 
		    MatchGroup=#match_group{vals=Acc}, SuccMap) ->
  NewClause=remove_succ_match(Match, BClause),
  {Succ, Hash1}=adapt(NewClause, Hash),
  make_successor_list(Rest, Hash1, BClause,MatchGroup#match_group{vals=[Val|Acc]},
		      gb_trees:insert(Val, Succ, SuccMap));
make_successor_list([], Hash, _BClause, MatchGroup, SuccMap) ->
  {SuccMap, MatchGroup, Hash}.

remove_all_fail_match(MatchSet, BClause) ->
  MatchList=gb_sets:to_list(MatchSet),
  lists:foldl(fun remove_fail_match/2, BClause, MatchList).

%%------------------------------------------------------------------------
%%
%% The following remove functions are used for size pruning
%%
%% remove_tests removes size tests which are true if a certain size test is
%% true
%%
%% remove_clauses removes clauses that contain a size test that is false
%% if a certain size test is false
%%
%%------------------------------------------------------------------------- 


prune_compatible_size(BClause=#b_clause{segments=BinSegs,next_clause=Next}, Size) ->
  case remove_tests(BinSegs, Size, []) of
    fail ->
      prune_compatible_size(Next, Size);
    NewSegs ->
      BClause#b_clause{segments=NewSegs, next_clause=prune_compatible_size(Next, Size)}
  end;

prune_compatible_size([], _Size) ->
  [].


remove_tests([SizeExpr=#size{all=AlliTerm}|Rest], Size=#size{all=AllTerm}, Acc) ->
  Alli = cerl:concrete(AlliTerm),
  All = cerl:concrete(AllTerm),
  Eq = false,
  Geq = all,
  case {All, Alli} of
    {Geq, Geq} ->
      case statically_bigger_or_equal(Size, SizeExpr) of
	true ->
	  remove_tests(Rest, Size, Acc);
	false ->
	  remove_tests(Rest, Size, [SizeExpr|Acc])
      end;
    {Geq, Eq} ->
      case statically_bigger(Size, SizeExpr) of
	true ->
	  fail;
	false ->
	  remove_tests(Rest, Size, [SizeExpr|Acc])
      end;
    {Eq, Geq} ->
      case statically_bigger(Size, SizeExpr) of
	true ->
	  fail;
	false ->
	  case statically_bigger_or_equal(SizeExpr, Size) of
	    true ->
	       remove_tests(Rest, Size, Acc);
	    false ->
	      remove_tests(Rest, Size, [SizeExpr|Acc])
	  end
      end;
    {Eq, Eq} ->
      case statically_different(Size, SizeExpr) of
	true ->
	  fail;
	false ->
	  case statically_equal(Size, SizeExpr) of
	    true ->
	       remove_tests(Rest, Size, Acc);
	    false ->
	      remove_tests(Rest, Size, [SizeExpr|Acc])
	  end
      end
  end;

remove_tests([Seg|Rest], Size, Acc) ->
  remove_tests(Rest, Size, [Seg|Acc]);
remove_tests([], _Size, Acc) ->
  lists:reverse(Acc).

prune_incompatible_size(BClause=#b_clause{segments=BinSegs, next_clause=Next}, Size) ->
  case [X || X <- BinSegs, is_incompatible(X, Size)] of
    [] ->
      BClause#b_clause{next_clause=prune_incompatible_size(Next, Size)};
    _ ->
      prune_incompatible_size(Next, Size)
  end;
prune_incompatible_size([], _Size) ->
  [].

is_incompatible(SizeExpr=#size{all=AlliTerm}, Size=#size{all=AllTerm}) ->
  Alli = cerl:concrete(AlliTerm),
  All = cerl:concrete(AllTerm),
  Eq = false,
  Geq = all,
  case {All, Alli} of
    {Geq, _} ->
      statically_bigger_or_equal(SizeExpr, Size);
    {Eq, Eq} ->
      statically_equal(SizeExpr, Size);
    _ ->
      false
  end;
is_incompatible(_,_) ->
  false.


%%-----------------------------------------------------------------------------
%%
%% remove_seg removes prunes all similar read actions
%%
%%-----------------------------------------------------------------------------

remove_seg(BinSeg, BClause=#b_clause{segments=BinSegs, next_clause=Next}) ->
  NewBinSegs  = remove_same_seg(BinSegs, tag(BinSeg)),
  BClause#b_clause{segments=NewBinSegs, next_clause=remove_seg(BinSeg, Next)};

remove_seg(_BinSeg, []) ->
  [].

remove_same_seg([#read_seg{tag=Tag}|Rest], Tag) ->
  remove_same_seg(Rest, Tag);
remove_same_seg([BinSeg|Rest], Tag) ->
  [BinSeg|remove_same_seg(Rest, Tag)];
remove_same_seg([], _Tag) ->
  [].

%%------------------------------------------------------------------------
%%
%% The following remove functions are used for match pruning
%%
%% remove_succ_match removes matches which are true if a certain match is
%% true and clauses containing a match that is false 
%%
%% remove_fail_match removes clauses that contain a match that is false
%% if a certain match is false
%%
%%------------------------------------------------------------------------- 

remove_succ_match(Match, BClause=#b_clause{segments=BinSegs, next_clause=Next}) -> 
  case mismatch(Match, BinSegs) of
    true ->
      remove_succ_match(Match, Next);
    NewBinSegs ->
      BClause#b_clause{segments=NewBinSegs, next_clause=remove_succ_match(Match, Next)} 
  end;
remove_succ_match(_Match, []) ->
  [].
mismatch(Match, BinSegs) ->
  mismatch(Match,BinSegs, []).
mismatch(Match=#match{tag=Tag, val=Val1}, [#match{tag=Tag, val=Val2}|Rest], Acc) ->
  case cerl:concrete(Val1)==cerl:concrete(Val2) of
    true -> mismatch(Match, Rest, Acc);
    false -> true
  end;
mismatch(Match, [First|Rest], Acc) ->
  mismatch(Match, Rest, [First|Acc]);
mismatch(_Match, [], Acc) ->
  lists:reverse(Acc).

remove_fail_match(Match, BClause=#b_clause{segments=BinSegs, next_clause=Next}) -> 
  case match(Match, BinSegs) of
    true ->
      remove_fail_match(Match, Next);
    false ->
      BClause#b_clause{next_clause=remove_fail_match(Match, Next)} 
  end;

remove_fail_match(_Match, []) ->
  [].
match(Match=#match{tag=Tag, val=Val1}, [#match{tag=Tag, val=Val2}|Rest]) ->
  case cerl:concrete(Val1)==cerl:concrete(Val2) of
    true -> true;
    false -> match(Match, Rest)
  end;
match(Match, [_|Rest]) ->
  match(Match, Rest);
match(_Match, []) ->
  false.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% choose_binseg/1 will chose the binseg or match with the 
%% highest count from a list, if there are binsegs and matches 
%% with the same count a binseg is chosen
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



choose_binseg(BClause) ->
  CountTree=simpl_count(BClause),
  BinSegs=b_clause_segments(BClause),
  Type = "uses_match",
  get_largest(BinSegs, BClause, CountTree, Type).

always_read(BinSegs, CountTree, BClause) ->
  No=no_of_clauses(BClause),
  find_binseg(BinSegs, CountTree, No).
    
find_binseg([BinSeg=#read_seg{}|Rest], CountTree, No) ->
  Tag=tag(BinSeg),
  case gb_trees:lookup(Tag, CountTree) of
    {value, {No, _, _, _}} ->
      {true, BinSeg};
    _ ->
      find_binseg(Rest, CountTree, No)
  end;

find_binseg([_|Rest], CountTree, No) ->
  find_binseg(Rest, CountTree, No);
find_binseg([], _CountTree, _No) ->
  false.

no_of_clauses(#b_clause{next_clause=Next}) ->
  1+no_of_clauses(Next);
no_of_clauses([]) ->
  0.


get_largest(BinSegs=[FirstSeg|_], BClause, CountTree, "left") ->
  Seg=FirstSeg,
    case always_read(BinSegs, CountTree, BClause) of
      {true, BinSeg} -> BinSeg;
      false -> FirstSeg
    end,
  legalize(Seg, BinSegs);

get_largest(BinSegs, BClause, CountTree, Type) ->
  Seg=
    case always_read(BinSegs, CountTree, BClause) of
      {true, BinSeg} -> BinSeg;
      false ->
	get_largest_count(BinSegs, CountTree, Type, {none, 0})
    end,  
  legalize(Seg, BinSegs).

get_largest_count([BinSeg|Rest], CountTree, "indexness", Top={_,Count}) -> 
  case BinSeg of
    #match{} ->
      Tag=tag(BinSeg),
      {value, {_, _, IC, _}} = gb_trees:lookup(Tag, CountTree),
      case IC >= Count of
	true ->
	  get_largest_count(Rest, CountTree, "indexness", {BinSeg, IC});
	false ->
	  get_largest_count(Rest, CountTree, "indexness", Top)
      end;
    #read_seg{} ->
      case Count of
	0 ->
	  get_largest_count(Rest, CountTree, "indexness", {BinSeg, 1});
	_ ->
	  get_largest_count(Rest, CountTree, "indexness", Top)
      end; 
    #size{} ->
      {value, {_, _, IC, _}} = gb_trees:lookup(BinSeg, CountTree),
      case IC >= Count of
	true ->
	  get_largest_count(Rest, CountTree, "indexness", {BinSeg, IC});
	false ->
	  get_largest_count(Rest, CountTree, "indexness", Top)
      end
  end;

get_largest_count([BinSeg|Rest], CountTree, "uses_match", Top={_,Count}) -> 
  case BinSeg of
    #match{} ->
      Tag=tag(BinSeg),
      {value, {_, MC, _, _}} = gb_trees:lookup(Tag, CountTree),
      case MC >= Count of
	true ->
	  get_largest_count(Rest, CountTree, "uses_match", {BinSeg, MC});
	false ->
	  get_largest_count(Rest, CountTree, "uses_match", Top)
      end;
    #read_seg{} ->
      case Count of
	0 ->
	  get_largest_count(Rest, CountTree, "uses_match", {BinSeg, 1});
	_ ->
	  get_largest_count(Rest, CountTree, "uses_match", Top)
      end;
    #size{} ->
      {value, {_, MC, _, _}} = gb_trees:lookup(BinSeg, CountTree),
      case MC >= Count of
	true ->
	  get_largest_count(Rest, CountTree, "uses_match", {BinSeg, MC});
	false ->
	  get_largest_count(Rest, CountTree, "uses_match", Top)
      end
  end;

get_largest_count([BinSeg|Rest], CountTree, "uses_read", Top={_,Count}) -> 
 
  case BinSeg of
    #match{} ->
      Tag=tag(BinSeg),
      {value, {_, MC, _, _}} = gb_trees:lookup(Tag, CountTree),
      NewMC = 1+MC/1000,
      case NewMC>Count of
	true ->
	  get_largest_count(Rest, CountTree, "uses_read", {BinSeg, NewMC});
	false ->
	  get_largest_count(Rest, CountTree, "uses_read", Top)
      end;
    #read_seg{} ->
      Tag=tag(BinSeg),
      {value, {RC, _, _, _}} = gb_trees:lookup(Tag, CountTree),
      case RC > Count of
	true ->
	  get_largest_count(Rest, CountTree, "uses_read", {BinSeg, RC});
	false ->
	  get_largest_count(Rest, CountTree, "uses_read", Top)
      end;
    #size{} ->
      {value, {_, MC, _, _}} = gb_trees:lookup(BinSeg, CountTree),
      NewMC = 1+MC/1000,
      case NewMC>Count of
	true ->
	  get_largest_count(Rest, CountTree, "uses_read", {BinSeg, NewMC});
	false ->
	  get_largest_count(Rest, CountTree, "uses_read", Top)
      end
  end;

get_largest_count([], _CountTree, _, {BinSeg, _Count}) ->
  BinSeg.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Utility functions
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Simpl count creates a mapping between segments and different 
% frequency valued used to synthesize the traversal order
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
simpl_count(BinClauses) ->
  simpl_count(BinClauses, gb_trees:empty()).

simpl_count(#b_clause{segments=BinSegs, next_clause=Next}, Tree) ->
  NewTree = simpl_count_segs(BinSegs, Tree),
  simpl_count(Next, NewTree);

simpl_count([], Tree) ->
  Tree.

simpl_count_segs([Seg=#match{}|Rest], Tree) ->
  Tag = match_tag(Seg),
  Val = match_val(Seg),
  case gb_trees:lookup(Tag, Tree) of
    {value, {SC, MC, IC, Set}} ->
      case gb_sets:is_member(Val, Set) of
	false ->
	  simpl_count_segs(Rest, gb_trees:update(Tag, {SC, MC+1, IC+1, gb_sets:add(Val, Set)}, Tree));
	true ->
	  simpl_count_segs(Rest, gb_trees:update(Tag, {SC, MC+1, IC, Set}, Tree))
      end;
    none ->
      simpl_count_segs(Rest, gb_trees:insert(match_tag(Seg), {0,1,1,gb_sets:singleton(Val)}, Tree))
  end;

simpl_count_segs([Seg=#read_seg{}|Rest], Tree) ->
  Tag = read_seg_tag(Seg),
  case gb_trees:lookup(Tag, Tree) of
    {value, {SC, MC, IC, Set}} ->
      simpl_count_segs(Rest, gb_trees:update(Tag, {SC+1, MC, IC, Set}, Tree));
    none ->
      simpl_count_segs(Rest, gb_trees:insert(Tag, {1,0,0,gb_sets:empty()}, Tree))
  end;

simpl_count_segs([Seg=#size{}|Rest], Tree) ->
  case gb_trees:lookup(Seg, Tree) of
    {value, {SC, MC, IC, Set}} ->
      simpl_count_segs(Rest, gb_trees:update(Seg, {SC, MC+1, IC, Set}, Tree));
    none ->
      simpl_count_segs(Rest, gb_trees:insert(Seg, {0,1,1,gb_sets:empty()}, Tree))
  end;

simpl_count_segs([], Tree) ->
  Tree.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% legalize makes sure that a binseg selection is legal, if it is not, a legal 
%% alternative is returned
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


legalize(BinSeg=#match{tag=Tag}, BinSegs) ->
  case has_read(Tag, BinSegs) of
    {true, ReadSeg} ->
      legalize(ReadSeg, BinSegs);
    false ->
      BinSeg
  end;

legalize(BinSeg=#read_seg{}, BinSegs) ->
  case size_precedes_read(BinSegs, BinSeg) of
    {true, Size} ->
      legalize(Size, BinSegs);
    false ->
      BinSeg
  end;
legalize(BinSeg=#size{}, BinSegs) ->
  case size_allowed(BinSegs, BinSeg) of
    {false, Read} ->
      legalize(Read, BinSegs);
    true ->
      BinSeg
  end.

has_read(Tag, [BinSeg=#read_seg{tag=Tag}|_]) ->
  {true,BinSeg};
has_read(Tag, [_|Rest]) ->
  has_read(Tag, Rest);
has_read(_, []) ->
  false.

size_precedes_read([Size=#size{}|_], _BinSeg) ->
  {true, Size};
size_precedes_read([BinSeg|_], BinSeg) ->
  false;
size_precedes_read([_|Rest], BinSeg) ->
  size_precedes_read(Rest, BinSeg).

size_allowed([#read_seg{}=Read|Rest], Size) -> 
  case size_need_read(Size, Read) of
    false ->
      size_allowed(Rest, Size);
    true ->
      {false, Read}
  end;
size_allowed([_|Rest], Size) ->
  size_allowed(Rest, Size);
size_allowed([], _Size) ->
  true.

size_need_read(#size{def_vars=Defs}, #read_seg{tag=Tag}) ->
  case [X || {tag, X} <- Defs, X==Tag] of
    [] ->
      false;
    _ ->
      true
  end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% combine(Size, Unit) returns a new size value either Size*Unit
%% or if Size is not an integer {Size, Unit}
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
combine(Size, Unit) ->
  case cerl:is_c_int(Size) of
    true ->
      cerl:c_int(cerl:int_val(Size)*cerl:int_val(Unit));
    false ->
      {Size, Unit}
  end.

calc_offset({{tag, Tag}, Unit}, SizeExpr=#size{def_vars=DefVars}) ->
  SizeExpr#size{def_vars=[{{tag,Tag}, Unit}|DefVars]};

calc_offset({Size, Unit}, SizeExpr=#size{vars=Vars}) ->
  case cerl:is_c_atom(Size) of
    true ->
      all = cerl:atom_val(Size),
      SizeExpr#size{all=Size};
    false ->
      true = cerl:is_c_var(Size),
      SizeExpr#size{vars=[{Size, Unit}|Vars]}
  end;

calc_offset(Size, SizeExpr=#size{const=Const}) ->
  SizeExpr#size{const=cerl:c_int(cerl:int_val(Size)+cerl:int_val(Const))}.
      
seg_type(Segment) ->      
  Type=cerl:bitstr_type(Segment),   
  Flags=cerl:bitstr_flags(Segment),
  {Type, Flags}.

statically_equal(Size1=#size{const=Const},
		 Size2=#size{const=Const}) ->
  Var1=size_all_vars(Size1),
  Var2=size_all_vars(Size2),
  same_set(Var1, Var2);
statically_equal(_,_) ->
  false.

statically_bigger(Size1=#size{const=Const1}, Size2=#size{const=Const2}) when Const1 > Const2 ->
  Var1=size_all_vars(Size1),
  Var2=size_all_vars(Size2),
  subset(Var2, Var1);
statically_bigger(_,_) ->
  false.

statically_bigger_or_equal(Size1=#size{const=Const1},Size2=#size{const=Const2}) when Const1 >= Const2 ->
  Var1=size_all_vars(Size1),
  Var2=size_all_vars(Size2),
  subset(Var2, Var1);
statically_bigger_or_equal(_,_) ->
  false.

statically_different(Size1, Size2) ->
  statically_bigger(Size1, Size2) or statically_bigger(Size2, Size1).

same_set([First|Rest], List) ->
  case lists:member(First, List) of
    true ->
      same_set(Rest, lists:delete(First, List));
    false ->
      false
  end;

same_set([], []) ->
  true;
same_set([], _) ->
  false.

subset([First|Rest], List) ->
  case lists:member(First, List) of
    true ->
      subset(Rest, lists:delete(First, List));
    false ->
      false
  end;

subset([], _List) ->
  true.

tag(T=#read_seg{}) ->
  read_seg_tag(T);
tag(T=#match{}) ->
  match_tag(T);
tag(#size{}) -> 
  [].

%% @spec instr_type(instr()) -> atom()
%% @doc returns the atom describing the instruction type

instr_type(Instr) ->
  case Instr of
    #read_seg{} ->
      read_seg;
    #bin_guard{} ->
      bin_guard;
    #size{} ->
      size;
    #match{} ->
      match;
    #label{} ->
      label;
    #goto{} ->
      goto;
    #match_group{} ->
      match_group
  end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% @spec pp(binary_match()|clause_tree())-> atom()
%% @doc pretty prints a clause_tree or the clause_tree of a binary_match
%% @end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-ifdef(BIN_PMATCH_DEBUG).
pp(#binary_match{clause_tree=CT}) ->
  pp(CT);

pp(#clause_tree{instr=I, success=Succ, fail=Fail}) -> 
  case I of
    #label{name=X} ->
      Int=cerl:concrete(X),
      io:format("Label: ~w~n", [Int]),
      pp(Succ);
    #size{vars=Vars, const=X, all=C_atom} ->
      Sign =
      case cerl:concrete(C_atom) of
	all ->
	  ">=";
	false ->
	  "="
      end,
      Int=cerl:concrete(X),
      io:format("Case size ~s ~w + ~w ~n", [Sign, Int, Vars]),
      do_split(Succ, Fail);
    #match{val=Val, tag=Tag} ->
      Cval= cerl:concrete(Val),
      io:format("Case ~w = r~w~n", [Cval, Tag]),
      do_split(Succ, Fail);
    #read_seg{tag=Tag, offset=Off, size=Size} ->
      Sz = 
	case cerl:is_c_int(Size) of
	  true -> cerl:concrete(Size);
	  false -> Size
	end,
      #size{vars=Vars, const=X} = Off,
      Int=cerl:concrete(X),
      io:format("r~w = read(~w + ~w, ~w)~n", [Tag, Vars, Int, Sz]),
      Lbl = get_label(Succ),
      io:format("goto ~w~n~n", [Lbl]),
      pp(Succ);
    #bin_guard{label=Lbl} ->
      io:format("Matched pattern: ~w~n~n", [Lbl]);
    #goto{} ->
      ok
  end.

do_split(Succ, Fail) ->
  SL = get_label(Succ),
  io:format("true -> ~w~n", [SL]),
  case get_label(Fail) of
    fail ->
      io:format("fail -> fail~n~n", []),
      pp(Succ);
    FL ->
      io:format("fail -> ~w~n~n", [FL]),
      pp(Succ),
      pp(Fail)
  end.

get_label(#clause_tree{instr=I}) ->
  case I of
    #label{name=X} ->
      cerl:concrete(X);
    #goto{label=X} ->
      cerl:concrete(X)
  end;
get_label([]) ->
  fail.
-endif.
%% @clear  No warnings if not ifdef(BIN_PMATCH_DEBUG)

% %% @spec full_size(binary_match())-> integer()
% %% @doc measures the number of nodes of the clause_tree in 
% %% a binary_match
% full_size(#binary_match{clause_tree=CT}) ->
%     full_size(CT, 0) + 1.
% full_size(#clause_tree{instr=I, success=Succ, fail=Fail}, Acc) ->
%     Acc1 = 
% 	case I of 
% 	    #label{name=X} ->
% 		max(cerl:concrete(X), Acc);
% 	    _ ->
% 		Acc
% 	end,
%     Acc2 = full_size(Succ, Acc1),
%     full_size(Fail, Acc2);

% full_size([], Acc) ->
%     Acc.

% max(A, B) when A>B ->
%     A;
% max(_A,B) ->
%     B.

% %% @spec a_height(binary_match())-> float()
% %% @doc measures the average height of the clause_tree in 
% %% a binary_match
% a_height(#binary_match{clause_tree=CT}) ->
%     Heights=a_height(CT, CT, [], 0),
%     Sorted=lists:keysort(2,Heights),
%     Added=calculate(Sorted),
%     lists:sum(Added)/length(Added).

% %% @spec max_height(binary_match())-> integer() 
% %% @doc measures the maximum height of the clause_tree in 
% %% a binary_match
% max_height(#binary_match{clause_tree=CT}) ->
%     Heights=a_height(CT, CT, [], 0),
%     [{High, _}|_]=lists:reverse(lists:keysort(1,Heights)),
%     High.

% calculate([{X,Token}|Rest]) -> 
%     calculate(Rest, [X], Token, []).

% calculate([{X,Token}|Rest], Acc, Token, AccList) ->
%     calculate(Rest, [X|Acc], Token, AccList);
% calculate([{X,NewToken}|Rest], Acc, _Token, AccList) ->
%     Val=lists:sum(Acc)/length(Acc),
%     calculate(Rest, [X], NewToken, [Val|AccList]);
% calculate([], Acc, _Token, AccList) ->
%     Val=lists:sum(Acc)/length(Acc),
%     [Val|AccList].

% a_height(#clause_tree{instr=I,success=Succ, fail=Fail}, OT, AccList, Acc) ->
%     case I of
% 	#bin_guard{label=X} ->
% 	    [{Acc+1, X}|AccList];
% 	#label{} ->
% 	    a_height(Succ, OT, AccList, Acc);
% 	#goto{label=Label} ->
% 	    {value,OtherSucc}=get_succ(OT, Label),
% 	    a_height(OtherSucc, OT, AccList, Acc);
% 	_ ->
% 	    NewList=a_height(Succ, OT, AccList, Acc+1),
% 	    a_height(Fail, OT, NewList, Acc+1)
%     end;

% a_height([], _OT, AccList,_Acc) ->
%     AccList.

% get_succ(#clause_tree{instr=I,success=Succ, fail=Fail}, Label) ->
%     case I of
% 	#label{name=Label} ->
% 	    {value, Succ};
% 	_ ->
% 	    case get_succ(Succ, Label) of
% 		none ->
% 		    get_succ(Fail, Label);
% 		Val ->
% 		    Val
% 	    end
%     end;
% get_succ([], _Label) ->
%     none.


