%% -*- erlang-indent-level: 2 -*-
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% HiPE Intermediate Code
%% ====================================================================
%%  Filename : 	hipe_icode.erl
%%  Module   :	hipe_icode
%%  Purpose  :  Provide primops for the Icode data structure.
%%  History  :	1997-? Erik Johansson (happi@csd.uu.se): Created.
%%              2001-01-30 EJ (happi@csd.uu.se):
%%                             Apply, primop, guardop removed
%%              2003-03-15 ES (happi@acm.org):
%%                             Started commenting in Edoc.
%%                             Moved pretty printer to separate file.
%%
%% $Id: hipe_icode.erl,v 1.80 2004/05/18 16:19:51 richardc Exp $
%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%@doc
%% This module implements "Linear Icode" and Icode instructions. 
%%          
%% <p> Icode is a simple (in that it has few instructions) imperative
%% language, used as the first Intermediate Code in the HiPE compiler.
%% Icode is closely related to Erlang, and Icode instructions operate
%% on Erlang terms. </p>
%%
%% <h2><a href="#type-icode">Icode</a></h2>
%%
%% <p> Linear Icode for a function consists of:
%%     <ul>
%%       <li> the function's name (<code>{M,F,A}</code>), </li>
%%       <li> a list of parameters, </li>
%%       <li> a list of instructions, </li>
%%       <li> data, </li>
%%       <li> information about whether the function is a leaf function, </li>
%%       <li> information about whether the function is a closure, and </li>
%%       <li> the range for labels and variables in the code. </li>
%%     </ul>
%% </p>
%%
%% <h2><a href="#type-icode_instruction">Icode Instructions</a> (and
%% their components)</h2>
%%
%% Control flow:
%% <dl>
%%    <dt><code><a href="#type-if">'if'</a> 
%%          {Cond::<a href="#type-cond">cond()</a>, 
%%           Args::[<a href="#type-arg">arg()</a>],
%%           TrueLabel::<a href="#type-label_name">label_name()</a>, 
%%           FalseLabel::<a href="#type-label_name">label_name()</a>
%%          } :: 
%%           <a href="#type-icode_instruction">icode_instruction()</a></code></dt>
%%    <dd>
%%        The if instruction compares the arguments (Args) with
%%        condition (Cond) and jumps to either TrueLabel or
%%        FalseLabel. (At the moment...) There are only binary
%%        conditions so the number of arguments should be two.
%%        <p>
%%        An if instructions ends a basic block and should be followed
%%        by a label (or be the last instruction of the code).
%%        </p></dd>
%%
%%    <dt><code><a href="#type-switch_val">switch_val</a> 
%%                    {Arg::<a href="#type-arg">arg()</a>, 
%%                     FailLabel::<a href="#type-label_name">label_name()</a>, 
%%                     Length::integer(), 
%%                     Cases::[{<a href="#type-symbol">symbol()</a>,<a
%%                     href="#type-label_name">label_name()</a>}] %% }::
%%           <a href="#type-icode_instruction">icode_instruction()</a></code></dt>
%%    <dd>
%%        The switch_val instruction compares the argument Arg to the
%%        symbols in the lists Cases, control is transfered to the label
%%        that corresponds to the first symbol that matches.  If no
%%        symbol matches control is transfered to FailLabel.  (NOTE: The
%%        length argument is not currently in use.)
%%        <p>
%%        The switch_val instruction can be assumed to be implemented as
%%        efficiently as possible given the symbols in the case
%%        list. (Jump-table, bianry-serach, or nested ifs)
%%        </p><p>
%%        A switch_val instructions ends a basic block and should be
%%        followed by a label (or be the last instruction of the code).
%%        </p></dd>
%%
%%    <dt><code><a href="#type-switch_tuple_arity">switch_tuple_arity</a>
%%         {
%%          Arg::<a href="#type-arg">arg()</a>, 
%%          FailLabel::<a href="#type-label_name">label_name()</a>, 
%%          Length::integer(),  
%%          Cases::[{integer(),<a href="#type-label_name">label_name()</a>}]
%%        }::
%%           <a href="#type-icode_instruction">icode_instruction()</a></code></dt>
%%    <dd>
%%        The switch_tuple_arity instruction compares the size of the
%%        tuple in the argument Arg to the integers in the lists Cases,
%%        control is transfered to the label that corresponds to the
%%        first integer that matches.  If no integer matches control is
%%        transfered to FailLabel.  (NOTE: The length argument is not
%%        currently in use.)
%%        <p>
%%        The switch_tuple_arity instruction can be assumed to be
%%        implemented as efficently as possible given the symbols in the
%%        case list. (Jump-table, bianry-serach, or nested ifs)
%%        </p><p>
%%        A switch_tuple_arity instructions ends a basic block and
%%        should be followed by a label (or be the last instruction of
%%        the code).
%%        </p></dd>
%%
%%    <dt><code>type {typ_expr, arg, true_label, false_label}}</code></dt>
%%    <dt><code>goto {label}</code></dt>
%%    <dt><code>label {name}</code></dt>
%% </dl>
%%
%% Moves:
%% <dl>
%%    <dt><code>move {dst, src}</code></dt>
%%    <dt><code>fmove {dst, src}</code></dt>
%%    <dt><code>phi {dst, arglist}</code></dt>
%% </dl>
%%
%% Function application:
%% <dl>
%%    <dt><code>call {[dst], fun, [arg], type, continuation, fail,
%%                    in_guard}</code></dt>
%%    <dd>
%%        Where <code>type</code> is one of {<code>'local', 'remote',
%%        'primop'</code>} and <code>in_guard</code> is either
%%        <code>'true'</code> or <code>'false'</code>.</dd>
%%    <dt><code>enter {fun, [arg], type}</code></dt>
%%    <dd>
%%        Where <code>type</code> is one of {<code>'local', 'remote',
%%        'primop'</code>} and <code>in_guard</code> is either
%%        <code>'true'</code> or <code>'false'</code>.</dd>
%%    <dt><code>return {[var]}</code></dt>
%%    <dd>
%%        <strong>WARNING:</strong> Multiple return values are yet not
%%        fully implemented and tested.
%%    </dd>
%% </dl>
%%
%% Error handling:
%% <dl>
%%    <dt><code>pushcatch {label, successor, type}</code></dt>
%%    <dt><code>restore_catch {reason_dst,type_dst,type, label}</code></dt>
%%    <dt><code>remove_catch {label}</code></dt>
%%    <dt><code>fail{[reason], type}</code></dt>
%%    <dd>
%%        Where <code>type</code> is one of {<code>'exit', 'throw',
%%        'fault', 'fault2', 'raise'</code>}.  For <code>fault2
%%        reason</code> is reason and trace.  For <code>raise
%%        [reason]</code> is <code>[Type,Reason]</code></dd>
%% </dl>
%%
%% Comments:
%% <dl>
%%    <dt><code>comment{Text::string()}</code></dt>
%% </dl>
%%
%% <h4>Notes</h4>
%%
%%  <p> A constant can only show up on the RHS of a `move' instruction
%%      and in `if' and `switch_*'</p>
%%  <p>
%%      Classification of primops should be like this:
%%      <ul>
%%      <li> <code>erlang:exit/1, erlang:throw/1, erlang:fault/1</code>, and 
%%           <code>erlang:fault/2</code> should use the
%%           <a href="#type-fail">fail-instruction</a> in Icode.</li>
%%      <li> Calls or tail-recursive calls to BIFs, operators, or interna
%%           functions should be implemented with `call' or `enter' 
%%           respectively, with the primop flag set.</li>
%%      <li> All other Erlang functions should be implemented with `call'
%%           or `enter' respectively, without the primop flag set.</li>
%%      </ul>
%%  </p>
%%
%% <h4>Primops</h4>
%%
%% <pre>
%%  Constructors:
%%    cons                       - [Car, Cdr]
%%    mktuple                    - [Element1, Element2, ..., ElementN]
%%    call_fun                   - [BoundArg1, ..., BoundArgN, Fun]
%%    enter_fun                  - [BoundArg1, ..., BoundArgN, Fun]
%%    {mkfun,MFA,MagicNum,Index} - [FreeVar1, FreeVar2, ..., FreeVarN]
%%
%%  Binaries:
%%    bs_init
%%    {bs_put_string, Bytes, Size}
%%    bs_final
%%
%%  Selectors:
%%    element              - [Index, Tuple]
%%    unsafe_hd            - [List]
%%    unsafe_tl            - [List]
%%    {unsafe_element, N}  - [Tuple], N:integer
%%    {unsafe_update_element, N}  - [Tuple, Val], N:integer
%%    {closure_element, N} - [Fun], N:integer
%%
%%  Arithmetic:       [Arg1, Arg2]
%%    '+','-','*','/','div','rem',
%%    'band','bor','bxor','bnot'
%%    'bsl','bsr',
%%
%%  Receive:         
%%    check_get_msg - []
%%    next_msg      - []
%%    select_msg    - []
%%    set_timeout   - [Timeout]
%%    clear_timeout - []	 %% stupid name - only resets message pointer
%%    suspend_msg   - []
%%
%% </pre>
%%
%% <h4>Guardops: (primops that can be used in guards and can fail.)</h4>
%%  <pre>
%%  Selectors:
%%    element - [Index, Tuple]
%%    unsafe_hd -[List]
%%    unsafe_tl -[List]
%%    {unsafe_element, N} - [Tuple], N:integer
%%
%%  Arithmetic:       [Arg1, Arg2]
%%    '+','-','*','/','div','rem',
%%   'band','bor','bxor','bnot'
%%   'bsl','bsr',
%%    fix_add, fix_sub               %% Do these exist?
%%
%%  Concurrency:
%%    {erlang,self,0}          - [] 
%% </pre>
%%
%%
%% <h4>Relational Operations (Cond in if instruction)</h4>
%% <pre>
%%    gt, lt, geq, leq,
%%    eqeq, neq, exact_eqeq, exact_neq
%% </pre>
%%
%% <h4>Type tests</h4>
%% <pre>
%%    list
%%    nil
%%    cons
%%    tuple
%%    {tuple, N}
%%    atom
%%    {atom, Atom}
%%    constant
%%    number
%%    integer
%%    {integer, N}
%%    fixnum
%%    bignum
%%    float
%%    pid
%%    port
%%    reference
%%    binary
%%    function
%% </pre>
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%=====================================================================

-module(hipe_icode).
-include("../main/hipe.hrl").
-include("hipe_icode.hrl").

%% @type icode(Fun, Params, IsClosure, IsLeaf, Code, Data, VarRange,LabelRange)
%%           Fun = mfa()
%%           Params = [var()]
%%           IsClosure = bool()
%%           IsLeaf = bool()
%%           Code = [icode_instruction()]
%%           Data = data()
%%           VarRange = {integer(),integer()}
%%           LabelRange = {integer(),integer()}
%%
%% @type icode_instruction(I) 
%%   I = if() | switch_val() | switch_tuple_arity() | type() | goto() | label()
%%        | move() | fmove() | phi() | call() | enter() | return() 
%%        | pushcatch() | restore_catch() | remove_catch() | fail() | comment()
%%
%% @type if(Cond, Args, TrueLabel, FalseLabel)
%%    Cond = cond()
%%    Args = [arg()]
%%    TrueLabel = label_name()
%%    FalseLabel = label_name()
%%
%% @type switch_val(Arg, FailLabel, Length, Cases) 
%%    Arg = arg()
%%    FailLabel=label_name()
%%    Length = integer()
%%    Cases = [{symbol(),label_name()}]
%%
%% @type switch_tuple_arity(Arg, FailLabel, Length, Cases)
%%    Arg = arg()
%%    FailLabel = label_name()
%%    Length = integer()
%%    Cases = [{symbol(), label_name()}]
%%
%% @type type(TypeExpr, Arg, True_label, False_label)
%%    TypeExpr = type_typr()
%%    Args = [arg()]
%%    TrueLabel = label_name()
%%    FalseLabel = label_name()
%%
%% @type goto(Label) Label = label_name()
%%
%% @type label(Name) Name=label_name()
%%
%% @type move(Dst, Src) Dst = var() Src = arg()
%%
%% @type fmove(Dst, Src) Dst = fvar() Src = farg()
%%
%% @type phi(Dst, Id, Arglist) 
%%           Dst = var()|fvar()
%%           Id = var()|fvar()
%%           Arglist=[{Pred, Src}]
%%           Pred = label_name()
%%           Var = var | fvar() 
%%
%% @type call(Dst, Fun, Arg, Type, Continuation, InGuard)
%%                   Dst = [var()]
%%                   Fun = mfa() | primop() | closure() 
%%                   Arg = [var()]
%%                   Type = call_type()
%%                   Continuation = [] | label_name()
%%                   Fail = []  | label_name()
%%                   InGuard = bool()
%%
%% @type enter(Fun, Arg, Type)
%%                   Fun = mfa() | primop() | closure() 
%%                   Arg = [var()] 
%%                   Type = call_type()
%%
%% @type return (Vars) Vars = [var()]
%%
%% @type pushcatch(Fail, Successor, Type) 
%%           Fail = label_name() Successor = label_name() Type= exit_type()
%%
%% @type restore_catch(ReasonDst,TypeDst,Type, Label) 
%%           ReasonDst = var()
%%           TypeDst = var()
%%           Type = exit_type()
%%           Label  = label_name()
%%
%% @type remove_catch(Label) Label = label_name()
%%
%% @type fail(Reason,Type)
%%    Type = exit_type()
%%    Reason = fail_reason() | fault2_reason() | raise_reason()
%%
%% @type comment(Text) Text = string()

%% @type call_type()  = 'local' | 'remote' | 'primop'
%% @type exit_type() = 'exit' | 'throw' | 'fault' | 'fault2' | 'raise'
%% @type fault2_reason(Reason,Trace) = cons(var(),cons(var(),[]))
%% @type raise_reason(Type,Reason) = cons(var(),cons(var(),[]))
%% @type fail_reason(Reason) = cons(var(),[])
%% @type cond() = gt | lt | geq | leq | eqeq | neq | exact_eqeq | exact_neq
%% @type type_type() = 
%%      list
%%    | nil
%%    | cons
%%    | tuple
%%    | {tuple, integer()}
%%    | atom
%%    | {atom, atom()}
%%    | constant
%%    | number
%%    | integer
%%    | {integer, integer()}
%%    | fixnum
%%    | bignum
%%    | float
%%    | pid
%%    | port
%%    | reference
%%    | binary
%%    | function
%%
%% @type mfa(Mod,Fun,Arity) = {atom(),atom(),integer()}

%% @type arg() = var() | const()
%% @type farg() = fvar() | float()
%% @type var(Name) Name=integer()
%% @type fvar(Name) Name=integer()
%% @type label_name(Name) Name = integer()
%% @type symbol(S) = atom() | number()
%% @type const(C)  C = const_fun() | immediate()
%% @type const_fun(MFA,U,I,Args) = {MFA,U,I,Args}
%%    MFA = mfa()
%%    U = integer()
%%    I = integer()
%%    Args = [var()]
%% @type immediate(I) = I
%%    I = term()
%% @end


%% ____________________________________________________________________
%%
%% Exports
%%
-export([mk_icode/7, %% mk_icode(Fun, Params, IsClosure, IsLeaf, 
		     %%          Code, VarRange, LabelRange)
	 mk_icode/8, %% mk_icode(Fun, Params, IsClosure, IsLeaf, 
		     %%          Code, Data, VarRange, LabelRange)
	 mk_typed_icode/8,
	 icode_fun/1,
	 icode_params/1,
	 icode_params_update/2,
	 icode_is_closure/1,
	 icode_is_leaf/1,
	 icode_code/1,
	 icode_code_update/2,
	 icode_data/1,
	 %% icode_data_update/2,
	 icode_var_range/1,
	 icode_label_range/1,
	 icode_info/1,
	 icode_info_update/2]).

-export([mk_if/4,           %% mk_if(Op, Args, TrueLbl, FalseLbl)
	 %% mk_if/5,	    %% mk_if(Op, Args, TrueLbl, FalseLbl, Prob)
	 if_op/1,
	 if_true_label/1,
	 if_false_label/1,
	 if_args/1,
	 if_pred/1,
	 %% is_if/1,
	 
	 mk_switch_val/4,
	 %% mk_switch_val/5,
	 switch_val_arg/1,
	 switch_val_fail_label/1,
	 %% switch_val_length/1,
	 switch_val_cases/1,
	 switch_val_cases_update/2,
	 %% is_switch_val/1,
	 
	 mk_switch_tuple_arity/4,
	 %% mk_switch_tuple_arityl/5,
	 switch_tuple_arity_arg/1,
	 switch_tuple_arity_fail_label/1,
	 %% switch_tuple_arity_length/1,
	 switch_tuple_arity_cases/1,
	 switch_tuple_arity_cases_update/2,
	 %% is_switch_tuple_arity/1,

	 mk_type/4,        %% mk_type(Args, Type, TrueLbl, FalseLbl)
	 mk_type/5,	   %% mk_type(Args, Type, TrueLbl, FalseLbl, P)
	 type_args/1,
	 %% type_args_update/2,
	 type_type/1,
	 type_true_label/1,
	 type_false_label/1,
	 type_pred/1,
	 %% is_type/1,

	 mk_guardop/5,     %% mk_guardop(Dst, Fun, Args,  Continuation, Fail)
	 mk_primop/3,      %% mk_primop(Dst, Fun, Args)
	 mk_primop/5,      %% mk_primop(Dst, Fun, Args, Cont, Fail)
	 mk_typed_call/6,  %% mk_call(Dst, Mod, Fun, Args, Type, DstType)
	 mk_call/5,	   %% mk_call(Dst, Mod, Fun, Args, Type)
	 %% mk_call/7,	   %% mk_call(Dst, Mod, Fun, Args, Type,
	                   %%         Continuation, Fail)
	 mk_call/8,	   %% mk_call(Dst, Mod, Fun, Args, Type,
	                   %%         Continuation, Fail, Guard)
	 call_dstlist/1,
	 call_dstlist_update/2,
	 call_dst_type/1,
	 call_args/1,
	 call_args_update/2,
	 call_fun/1,
	 call_fun_update/2,
	 call_type/1,
	 call_continuation/1,
	 call_fail/1,
	 call_set_fail/2,
	 call_set_continuation/2,
	 is_call/1, 
	 call_in_guard/1,
	 call_fails/1,

	 mk_goto/1,              %% mk_goto(Lbl)
	 goto_label/1,
	 
	 mk_enter/4,             %% mk_enter(Mod, Fun, Args, Type)
	 mk_enter_primop/2,      %% mk_enter(Op, Type)
	 enter_fun/1,
	 enter_args/1,
	 enter_type/1,
	 %% is_enter/1,
	 
	 mk_fmove/2,             %% mk_fmove(Dst, Src)

	 mk_return/1,            %% mk_return(Vars)
	 %% mk_fail/1,	         %% mk_fail(Reason) type = exit
	 mk_fail/2,              %% mk_fail(Reason, Type)
	 mk_move/2,              %% mk_move(Dst, Src)
	 mk_moves/2,             %% mk_moves(DstList, SrcList)
	 mk_pushcatch/2,         %% mk_pushcatch(Label, Successor)
	 mk_pushtry/2,           %% mk_pushtry(Label, Successor)
	 mk_restore_catch/2,     %% mk_restore_catch(ReasonDst, Label)
	 mk_restore_try/3,       %% mk_restore_try(ReasonDst, TypeDst, Label)
	 mk_remove_catch/1,      %% mk_remove_catch(Label)
	 %% mk_elements/2,       %% mk_elements(Tuple, Vars)
	 mk_label/1,             %% mk_label(Name)
	 mk_new_label/0,         %% mk_new_label()
	 mk_comment/1,           %% mk_comment(Text)
	 mk_const/1,             %% mk_const(Const)
	 %% mk_const_fun/4,	 %% mk_const_fun(MFA,U,I,Args)
	 mk_var/1,               %% mk_var(Id)
	 annotate_var/2,         %% annotate_var(Var, Type)
	 mk_reg/1,               %% mk_reg(Id)
	 mk_fvar/1,              %% mk_fvar(Id)
	 mk_new_var/0,           %% mk_new_var()
	 mk_new_fvar/0,          %% mk_new_fvar()
	 mk_new_reg/0,           %% mk_new_reg()
	 mk_phi/1,               %% mk_phi(Id)
	 mk_phi/2                %% mk_phi(Id, ArgList)
	]).

%%
%% Identifiers
%%

-export([type/1,
	 %% is_fail/1,
	 %% is_return/1,
	 is_move/1,
	 %% is_pushcatch/1,
	 is_restore_catch/1,
	 %% is_remove_catch/1,
	 is_goto/1,
	 is_label/1,
	 is_comment/1,
	 is_const/1,
	 is_const_fun/1,
	 is_var/1,
	 %% is_annotated_var/1,
	 is_fvar/1,
	 is_reg/1,
	 is_var_or_fvar_or_reg/1,
	 %% is_uncond/1,
	 %% is_fmove/1,
         is_phi/1]).

%%
%% Selectors
%%

-export([phi_dst/1,
         phi_id/1,
         %% phi_args/1,
         phi_arg/2,
         phi_arglist/1,
	 phi_enter_pred/3,
	 phi_remove_pred/2,
	 phi_redirect_pred/3,
	 move_dst/1,
	 move_src/1,
	 move_src_update/2,
	 pushcatch_label/1,
	 pushcatch_successor/1,
	 restore_catch_reason_dst/1,
	 restore_catch_type_dst/1,
	 restore_catch_type/1,
	 restore_catch_label/1,
	 remove_catch_label/1,
	 label_name/1,
	 comment_text/1,
	 return_vars/1,
	 fail_reason/1,
	 fail_type/1,
	 %% var_name/1,
	 %%var_annotation/1,
	 %% fvar_name/1,
	 %% reg_name/1,		 
	 const_value/1,
	 %% info/1,
	 fmove_dst/1,
	 fmove_src/1
	 %% fmove_src_update/2
	]).

%%
%% Misc
%%

-export([args/1,
	 uses/1,
	 defines/1,
	 is_safe/1,
	 fixup_fallthroughs/1,
	 strip_comments/1,
	 subst/2,
	 subst_uses/2,
	 subst_defines/2,
	 redirect_jmp/3,
	 successors/1,
	 is_branch/1,
	 is_leaf_code/1 % needs to be used in beam_to_icode
	]).

-export([highest_var/1,highest_label/1]).

%%---------------------------------------------------------------------
%% 
%% Icode
%%
%%---------------------------------------------------------------------

%% @spec mk_icode(Fun::mfa(), Params::[var()], Closure::bool(), 
%%                Leaf::bool(), Code::[icode_instruction()],
%%                VarRange::{integer(),integer()}, 
%%                LabelRange::{integer(),integer()}) -> icode()
%%
mk_icode(Fun, Params, Closure, Leaf, Code, VarRange, LabelRange) ->
  #icode{'fun'=Fun, params=Params, code=Code,
	 closure=Closure,
	 leaf=Leaf,
	 data=hipe_consttab:new(),
	 var_range=VarRange,
	 label_range=LabelRange}.
%% @spec mk_icode(Fun::mfa(), Params::[var()], Closure::bool(), Leaf::bool(), 
%%                Code::[icode_instruction()],  Data::data(),
%%                VarRange::{integer(),integer()}, 
%%                LabelRange::{integer(),integer()}) -> icode()
%%
mk_icode(Fun, Params, Closure, Leaf, Code, Data, VarRange, LabelRange) ->
  #icode{'fun'=Fun, params=Params, code=Code,
	 data=Data, closure=Closure, leaf=Leaf,
	 var_range=VarRange,
	 label_range=LabelRange}.
mk_typed_icode(Fun, Params, Closure, Leaf, Code, VarRange, 
	       LabelRange, ArgType) ->
  #icode{'fun'=Fun, 
	 params=Params,
	 code=Code,
	 closure=Closure,
	 leaf=Leaf,
	 data=hipe_consttab:new(),
	 var_range=VarRange,
	 label_range=LabelRange,
	 info=[{arg_type, ArgType}]}.
%% @spec icode_fun(I::icode()) -> mfa()
icode_fun(Icode) -> Icode#icode.'fun'.
%% @spec icode_params(I::icode()) -> [var()]
icode_params(Icode) -> Icode#icode.params.
%% @spec icode_params_update(I::icode(),[var()]) -> icode()
icode_params_update(Icode, Params) -> 
  Icode#icode{params=Params}.
%% @spec icode_is_closure(I::icode()) -> bool()
icode_is_closure(Icode) -> Icode#icode.closure.
%% @spec icode_is_leaf(I::icode()) -> bool()
icode_is_leaf(Icode) -> Icode#icode.leaf.
%% @spec icode_code(I::icode()) -> [icode_instruction()]
icode_code(Icode) -> Icode#icode.code.
%% @spec icode_code_update(I::icode(), [icode_instruction()]) -> icode()
icode_code_update(Icode, NewCode) -> 
  Vmax = highest_var(NewCode),
  Lmax = highest_label(NewCode),
  Icode#icode{code=NewCode, var_range={0,Vmax}, label_range={0,Lmax}}.
%% @spec icode_data(I::icode()) -> data()
icode_data(Icode) -> Icode#icode.data.
%% %% @spec icode_data_update(I::icode(),data()) -> icode()
%% icode_data_update(Icode, NewData) -> Icode#icode{data=NewData}.
icode_var_range(Icode) -> Icode#icode.var_range.
icode_label_range(Icode) -> Icode#icode.label_range.
icode_info(Icode) -> Icode#icode.info.
icode_info_update(Icode, Info) -> Icode#icode{info=Info}.

%% ____________________________________________________________________
%% Instructions
%%

%%
%% if
%%

mk_if(Op, Args, TrueLbl, FalseLbl) ->
  #'if'{op=Op, args=Args, true_label=TrueLbl, false_label=FalseLbl, p=0.5}.
%% mk_if(Op, Args, TrueLbl, FalseLbl, P) ->
%%   #'if'{op=Op, args=Args, true_label=TrueLbl, false_label=FalseLbl, p=P}.
if_op(Cond) -> Cond#'if'.op.
if_args(Cond) -> Cond#'if'.args.
if_true_label(Cond) -> Cond#'if'.true_label.
if_false_label(Cond) -> Cond#'if'.false_label.
if_pred(Cond) -> Cond#'if'.p.

%%
%% switch_val
%%

mk_switch_val(Arg, Fail_label, Length, Cases) ->
  #switch_val{arg=Arg, fail_label=Fail_label, length=Length, cases=Cases}.
switch_val_arg(Cond) -> Cond#switch_val.arg.
switch_val_fail_label(Cond) -> Cond#switch_val.fail_label.
%% switch_val_length(Cond) -> Cond#switch_val.length.
switch_val_cases(Cond) -> Cond#switch_val.cases.
switch_val_cases_update(Cond, NewCases) -> 
  Cond#switch_val{cases = NewCases}.
%% switch_val_info(Cond) -> Cond#switch_val.info.

%%
%% switch_tuple_arity
%%

mk_switch_tuple_arity(Arg, Fail_label, Length, Cases) ->
  #switch_tuple_arity{arg=Arg, fail_label=Fail_label, length=Length, cases=Cases}.
switch_tuple_arity_arg(Cond) -> Cond#switch_tuple_arity.arg.
switch_tuple_arity_fail_label(Cond) -> Cond#switch_tuple_arity.fail_label.
%% switch_tuple_arity_length(Cond) -> Cond#switch_tuple_arity.length.
switch_tuple_arity_cases(Cond) -> Cond#switch_tuple_arity.cases.
switch_tuple_arity_cases_update(Cond, NewCases) -> 
  Cond#switch_tuple_arity{cases = NewCases}.
%% switch_tuple_arity_info(Cond) -> Cond#switch_tuple_arity.info.

%%
%% type
%%

mk_type(X, Type, TrueLbl, FalseLbl) -> 
  #type{type=Type, args=X, true_label=TrueLbl, false_label=FalseLbl, p=0.5}.
mk_type(X, Type, TrueLbl, FalseLbl, P) -> 
  #type{type=Type, args=X, true_label=TrueLbl, false_label=FalseLbl, p=P}.
type_args(T) -> T#type.args.
%% type_args_update(T, Args) -> T#type{args=Args}.
type_type(T) -> T#type.type.
type_true_label(T) -> T#type.true_label.
type_false_label(T) -> T#type.false_label.
type_pred(T) -> T#type.p.

%%
%% goto
%%

mk_goto(Lbl) -> #goto{label=Lbl}.
goto_label(G) -> G#goto.label.
is_goto(#goto{}) -> true;
is_goto(_) -> false.

%%
%% return
%%

mk_return(Xs) -> #return{vars=Xs}.
return_vars(R) -> R#return.vars.

%%
%% fail
%%

%% mk_fail(Reason) when list(Reason) -> #fail{reason=Reason, type=exit}.
mk_fail(Reason, Type) when list(Reason) -> #fail{reason=Reason, type=Type}.
fail_reason(F) -> F#fail.reason.
fail_type(F) -> F#fail.type.

%%
%% move
%%

mk_move(X, Y) -> #move{dst=X, src=Y}.
move_dst(M) -> M#move.dst.
move_src(M) -> M#move.src.
move_src_update(M, NewSrc) -> M#move{src=NewSrc}.
is_move(#move{}) -> true;
is_move(_) -> false.

%%
%% phi
%%

%% The id field is not entirely redundant. It is used in mappings
%% in the SSA pass since the dst field can change.
mk_phi(Var) -> #phi{dst = Var, id = Var, arglist = []}.
mk_phi(Var, ArgList) -> #phi{dst = Var, id = Var, arglist = ArgList}.
phi_dst(P) -> P#phi.dst.
phi_id(P) -> P#phi.id.
phi_args(P) -> [X || {_,X} <- P#phi.arglist].
phi_arg(P, Pred) -> 
  case lists:keysearch(Pred, 1, P#phi.arglist) of
    false -> exit('No such predecessor to phi', {Pred, P});
    {value, {_, Var}} -> Var
  end.
phi_arglist(P) -> P#phi.arglist.
is_phi(#phi{}) -> true;
is_phi(_) -> false.
phi_enter_pred(P, Pred, Var) ->
  P#phi{arglist=[{Pred,Var}|lists:keydelete(Pred, 1, P#phi.arglist)]}.
phi_remove_pred(P, Pred) ->
  P#phi{arglist=lists:keydelete(Pred, 1, P#phi.arglist)}.
phi_argvar_subst(P, Subst) -> 
  NewArgList = [{Pred, subst1(Subst, Var)} || {Pred, Var}<-P#phi.arglist],
  P#phi{arglist=NewArgList}.
phi_redirect_pred(P, OldPred, NewPred)->
  Subst = [{OldPred, NewPred}],
  NewArgList = [{subst1(Subst, Pred), Var} || {Pred, Var}<-P#phi.arglist],
  P#phi{arglist=NewArgList}.

%%
%% primop and guardop
%%
%% Whether a function is a "primop" - i.e., an internal thing - or not,
%% is really only shown by its name.  An {M,F,A} always represents a
%% function in some Erlang module (althought it might be a BIF, and
%% could possibly be inline expanded).  It is convenient to let the
%% constructor functions check the name and set the type automatically,
%% especially for guardops - some guardops are primitives and some are
%% MFA:s, and this way we won't have to rewrite all calls to mk_guardop
%% to flag whether they are primops or not.

mk_primop(DstList, Fun, ArgList) ->
  mk_primop(DstList, Fun, ArgList, [], []).
mk_primop(DstList, Fun, ArgList, Continuation, Fail) ->
  mk_primop(DstList, Fun, ArgList, Continuation, Fail, false).
mk_primop(DstList, Fun, ArgList, Continuation, Fail, Guard) ->
  Type = find_call_type(Fun),
  make_call(DstList, Fun, ArgList, Type, Continuation, Fail, Guard).

%% Note that a 'guardop' is just a primop that occurred in a guard.

mk_guardop(DstList, Fun, ArgList, True, False) ->
  mk_primop(DstList, Fun, ArgList, True, False, true).

find_call_type(Fun) ->
  case Fun of
    {M,F,A} when atom(M), atom(F), integer(A) -> remote;
    {_,_,_} -> exit({bad_primop, Fun});
    _ -> primop
  end.

%%
%% call
%%
mk_typed_call(Dst, M, F, Args, Type, DstType) ->
  Call = mk_call(Dst, M, F, Args, Type),
  Call#call{dst_type=[DstType]}.

mk_call(DstList, M, F, ArgList, Type) ->
  mk_call(DstList, M, F, ArgList, Type, [], [], false).
%% mk_call(DstList, M, F, ArgList, Type, Continuation, Fail) ->
%%   mk_call(DstList, M, F, ArgList, Type, Continuation, Fail, false).
mk_call(DstList, M, F, ArgList, Type, Continuation, Fail, Guard)
  when atom(M), atom(F) ->
  Fun = {M,F,length(ArgList)},
  make_call(DstList, Fun, ArgList, Type, Continuation, Fail, Guard).

%% The common constructor for all calls
%%
%% Note: If the "guard" flag is `true', it means that if the call fails,
%% we can simply jump to the Fail label (if it exists) without
%% generating any additional exception information - it isn't needed.
%%
make_call(DstList, Fun, ArgList, Type, Continuation, Fail, Guard) ->
  case Type of
    local -> ok;
    remote -> ok;
    primop -> ok;
    _ -> exit({bad_call_type, Type})
  end,
  #call{dstlist=DstList, 'fun'=Fun, args=ArgList,
	type=Type, continuation_label=Continuation, fail_label=Fail,
	in_guard=Guard}.

call_dstlist(C) -> C#call.dstlist.
call_dstlist_update(C,Dest) -> C#call{dstlist=Dest}.
call_dst_type(C) -> C#call.dst_type.
%% @spec (C::call()) -> [arg()]
call_args(C) -> C#call.args.
%% @spec (C::call(), [arg()]) -> call()
call_args_update(C,Args) -> C#call{args=Args}.
call_fun(C) -> C#call.'fun'.
%% Note that updating the name field requires recomputing the call type,
%% in case it changes from a remote/local call to a primop call.
call_fun_update(C, Fun) ->
  Type = case find_call_type(Fun) of
	   primop -> primop;
	   _ -> C#call.type
	 end,
  C#call{'fun'=Fun, type=Type}.
call_type(C) -> C#call.type.
call_continuation(C) -> C#call.continuation_label.
call_fail(C) ->  C#call.fail_label.
call_set_continuation(I, Continuation) ->  
  I#call{continuation_label = Continuation}.
call_set_fail(I, Fail) ->
  I#call{fail_label = Fail}.
is_call(#call{}) -> true;
is_call(_) -> false.
call_in_guard(I) -> I#call.in_guard.
call_fails(I) ->
  case I#call.'fun' of
    cons -> false;
    closure_element -> false;
    fclearerror -> false;
    fp_add -> false;      
    fp_sub -> false;
    fp_mul -> false;
    fp_div -> false;
    mkfun -> false;
    next_msg -> false;
    unsafe_untag_float -> false;
    unsafe_tag_float -> false;
    unsafe_add -> false;
    unsafe_sub -> false;
    unsafe_bor -> false;
    unsafe_bxor -> false;
    unsafe_bnot -> false;
    unsafe_hd -> false;
    unsafe_tl -> false;
    {unsafe_element,_} -> false;
    {unsafe_update_element,_} -> false;
    _ -> true
  end.

%%
%% enter
%%

mk_enter(Mod, Fun, Args, Type) -> 
  case Type of
    local -> ok;
    remote -> ok;
    primop -> ok;
    _ -> exit({bad_enter_type, Type})
  end,
  #enter{'fun'={Mod,Fun,length(Args)}, args=Args, type=Type}.
enter_fun(E) -> E#enter.'fun'.
enter_args(E) -> E#enter.args.
enter_type(E) -> E#enter.type.
%%enter_fail(I) ->  I#enter.fail_label.
%%enter_set_fail(I, Fail) ->
%%  I#enter{fail_label = Fail}.

mk_enter_primop(Op, Args) ->
  #enter{'fun'=Op, args=Args, type=primop}.

%%
%% pushcatch
%%

mk_pushcatch(Label, Successor) ->
  #pushcatch{label=Label, successor=Successor}.
mk_pushtry(Label, Successor) ->
  #pushcatch{label=Label, successor=Successor, type='try'}.
pushcatch_label(P) -> P#pushcatch.label.
pushcatch_successor(P) -> P#pushcatch.successor.

%%
%% remove_catch
%%

mk_remove_catch(Label) -> #remove_catch{id = Label}.
remove_catch_label(P) -> P#remove_catch.id.

%%
%% restore_catch
%%

mk_restore_catch(ReasonDst, Label) ->
  #restore_catch{reason_dst=ReasonDst, type_dst=[],
		 catch_type='catch', id=Label}.
mk_restore_try(ReasonDst, TypeDst, Label) ->
  #restore_catch{reason_dst=ReasonDst, type_dst=TypeDst, 
		 catch_type='try', id=Label}.
restore_catch_reason_dst(P) -> P#restore_catch.reason_dst.
restore_catch_type_dst(P) -> P#restore_catch.type_dst.
restore_catch_type(P) -> P#restore_catch.catch_type.
restore_catch_label(P) -> P#restore_catch.id.
is_restore_catch(#restore_catch{}) -> true;
is_restore_catch(_) -> false.

%%
%% label
%%

mk_label(Name) -> #label{name=Name}.
label_name(L) -> L#label.name.
is_label(#label{}) ->true;
is_label(_) -> false.

%%
%% comment
%%

mk_comment(Txt) -> #comment{text=Txt}.
comment_text(C) -> C#comment.text.
%% @spec is_comment(Instr::icode_instruction()) -> bool()
%% @doc  True if this is the Icode instruction for comments.
is_comment(#comment{}) -> true;
is_comment(_) -> false.

%% ____________________________________________________________________
%% 

%%
%% Arguments (variables and constants)
%%

mk_const(C) -> {const,{flat,C}}.
%% mk_const_fun(MFA,U,I,Args) -> {const,{const_fun,{MFA,U,I,Args}}}.
const_value({const,{flat,X}}) -> X;
const_value({const,{const_fun,X}}) -> X.
%% @spec is_const(icode_arg()) -> bool()
is_const({const,_}) -> true;
is_const(_) -> false.
%% @spec is_const_fun(icode_arg()) -> bool()
is_const_fun({const,{const_fun,_}}) -> true;
is_const_fun(_) -> false.

mk_var(V) -> {var,V}.
var_name({var,Name}) -> Name.
%% @spec is_var(icode_arg()) -> bool()
is_var({var,_}) -> true;
is_var(_) -> false.
  
%% This representation of a variable is used only when pretty printing
%% typed icode.
annotate_var({var, Name}, Type)-> {var, Name, Type}.
%%is_annotated_var({var, _Name, _Type})-> true;
%%is_annotated_var(_) -> false.
%%var_annotation({var, _Name, Type})-> Type.
 
mk_reg(V) -> {reg,V}.
reg_name({reg,Name}) -> Name.
%% @spec is_reg(icode_arg()) -> bool()
is_reg({reg,_}) -> true;
is_reg(_) -> false.

%% @spec is_var_or_fvar_or_reg(icode_arg()) -> bool()
is_var_or_fvar_or_reg({var,_}) -> true;
is_var_or_fvar_or_reg({fvar,_}) -> true;
is_var_or_fvar_or_reg({reg,_}) -> true;
is_var_or_fvar_or_reg(_) -> false.

mk_fvar(V) -> {fvar,V}.
fvar_name({fvar,Name}) -> Name.
%% @spec is_fvar(icode_arg()) -> bool()
is_fvar({fvar,_}) -> true;
is_fvar(_) -> false.

%%
%% Floating point Icode instructions.
%%

%%
%% fmove
%%

mk_fmove(X, Y) -> #fmove{dst=X, src=Y}.
fmove_dst(M) -> M#fmove.dst.
fmove_src(M) -> M#fmove.src.
%% fmove_src_update(M, NewSrc) -> M#fmove{src=NewSrc}.

%%
%% info for all instructions
%%

%% info(I) ->
%%   case type(I) of
%%     'if' -> I#'if'.info;
%%     type -> I#type.info;
%%     goto -> I#goto.info;
%%     move -> I#move.info;
%%     fail -> I#fail.info;
%%     call -> I#call.info;
%%     enter -> I#enter.info;
%%     return -> I#return.info;
%%     pushcatch -> I#pushcatch.info;
%%     restore_catch -> I#restore_catch.info;
%%     remove_catch -> I#remove_catch.info;
%%     comment -> I#comment.info;
%%     switch_val -> I#switch_val.info;
%%     switch_tuple_arity -> I#switch_tuple_arity.info;
%%     label -> I#label.info;
%%     fmove -> I#fmove.info
%%   end.

%% info_add(Instr, Info) ->
%%   info_update(Instr, [Info | info(Instr)]).

%% info_update(I, NewInfo) ->
%%   case type(I) of 
%%     'if' -> I#'if'{info = NewInfo};
%%     type -> I#type{info = NewInfo};
%%     goto -> I#goto{info = NewInfo};
%%     move -> I#move{info = NewInfo};
%%     fail -> I#fail{info = NewInfo};
%%     call -> I#call{info = NewInfo};
%%     enter -> I#enter{info = NewInfo};
%%     return -> I#return{info = NewInfo};
%%     pushcatch -> I#pushcatch{info = NewInfo};
%%     restore_catch -> I#restore_catch{info = NewInfo};
%%     remove_catch -> I#remove_catch{info = NewInfo};
%%     comment -> I#comment{info = NewInfo};
%%     switch_val -> I#switch_val{info = NewInfo};
%%     switch_tuple_arity -> I#switch_tuple_arity{info = NewInfo};
%%     label -> I#label{info = NewInfo};
%%     fmove -> I#fmove{info = NewInfo}
%%   end.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Liveness info 
%%

%% @spec uses(icode_instruction()) -> [icode_arg()]
uses(Instr) ->
  remove_constants(args(Instr)).

%% @spec args(I::icode_instruction()) -> [var()]
args(I) ->
  case element(1, I) of
    'if' -> if_args(I);
    switch_val -> [switch_val_arg(I)];
    switch_tuple_arity -> [switch_tuple_arity_arg(I)];
    type -> type_args(I);
    move -> [move_src(I)];
    fail -> fail_reason(I);
    call -> I#call.args;
    enter -> I#enter.args;
    return -> I#return.vars;
    fmove -> [fmove_src(I)];
    phi -> phi_args(I);
    %%    goto -> [];
    %%    pushcatch -> [];
    %%    restore_catch -> [];
    %%    remove_catch -> [];
    %%    comment -> [];
    %%    label -> []
    _ -> []
  end.

defines(I) ->
  case element(1, I) of
    move -> remove_constants([I#move.dst]);
    fmove -> remove_constants([I#move.dst]);
    call -> remove_constants(I#call.dstlist);
    restore_catch -> 
      case I#restore_catch.catch_type of
	'try' ->
	  remove_constants([I#restore_catch.reason_dst,
			    I#restore_catch.type_dst]);
	'catch' ->
	  remove_constants([I#restore_catch.reason_dst])
      end;
    phi -> remove_constants([I#phi.dst]);
    %%    'if' -> [];
    %%    switch_val -> [];
    %%    switch_tuple_arity -> [];
    %%    type -> [];
    %%    goto -> [];
    %%    fail -> [];
    %%    enter -> [];
    %%    return -> [];
    %%    pushcatch -> [];
    %%    remove_catch -> [];
    %%    comment -> [];
    %%    label -> []
    _ -> []
  end.


remove_constants([]) ->
  [];
remove_constants([{const,_}|Xs]) ->
  remove_constants(Xs);
remove_constants([{reg,Var}|Xs]) ->
  [{reg,Var} | remove_constants(Xs)];
remove_constants([{var,Var}|Xs]) ->
  [{var,Var} | remove_constants(Xs)];
remove_constants([{fvar,Var}|Xs]) ->
  [{fvar,Var} | remove_constants(Xs)];
remove_constants([{colored,Var}|Xs]) ->
  [{colored,Var} | remove_constants(Xs)];
remove_constants([_|Xs]) ->
  remove_constants(Xs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Utilities
%%

%% @doc 
%%    Returns the type of the Icode instruction X. Hopefully, this can
%%    be compiled better than repeated calls to hipe_icode:is_branch/is_move/..
%% @end

type(I) when tuple(I) -> element(1, I);
type(O) -> ?EXIT({bad_icode_type,O}).

%%
%% Substitution: replace occurrences of X by Y if {X,Y} is in the
%%   Subst_list.

subst(Subst, X) ->
  subst_defines(Subst, subst_uses(Subst,X)).

subst_uses(Subst, X) ->
  case type(X) of
    'if' -> X#'if'{args = subst_list(Subst, X#'if'.args)};
    switch_val -> X#switch_val{arg = subst1(Subst, X#switch_val.arg)};
    switch_tuple_arity ->
      X#switch_tuple_arity{arg = subst1(Subst, X#switch_tuple_arity.arg)};
    type -> X#type{args = subst_list(Subst, X#type.args)};
    move -> X#move{src = subst1(Subst, X#move.src)};
    fail -> X#fail{reason = subst_list(Subst, fail_reason(X))};
    call -> X#call{args = subst_list(Subst, X#call.args)};
    enter -> X#enter{args = subst_list(Subst, X#enter.args)};
    return -> X#return{vars = subst_list(Subst, X#return.vars)};
    fmove -> X#fmove{src = subst1(Subst, X#fmove.src)};
    phi -> phi_argvar_subst(X, Subst);
    _ -> X
%%  goto -> X;
%%  pushcatch -> X;
%%  restore_catch -> X;
%%  remove_catch -> X;
%%  comment -> X;
%%  label -> X
  end.

subst_defines(Subst, X) ->
  case type(X) of
    move -> X#move{dst = subst1(Subst, X#move.dst)};
    call -> X#call{dstlist = subst_list(Subst, X#call.dstlist)};
    restore_catch -> 
      NX = X#restore_catch{reason_dst = subst1(Subst,
					       X#restore_catch.reason_dst)},
      NX#restore_catch{type_dst = subst1(Subst,NX#restore_catch.type_dst)};
    fmove -> X#fmove{dst = subst1(Subst, X#move.dst)};
    phi -> X#phi{dst = subst1(Subst, X#phi.dst)};
    _ -> X
  end.
%%    'if' -> X;
%%    switch_val -> X;
%%    switch_tuple_arity -> X;
%%    type -> X;
%%    goto -> X;
%%    fail -> X;
%%    enter -> X;
%%    return -> X;
%%    pushcatch -> X;
%%    remove_catch -> X;
%%    comment -> X;
%%    label -> X
%%  end.


subst_list(S,Xs) ->
  [subst1(S,X) || X <- Xs].

subst1([],X) -> X;
subst1([{X,Y}|_],X) -> Y;
subst1([_|Xs],X) -> subst1(Xs,X).

%%
%% @doc Returns the successors of an Icode branch instruction.
%%

successors(Jmp) ->
  case type(Jmp) of
    'if' -> [Jmp#'if'.true_label, Jmp#'if'.false_label];
    goto -> [Jmp#goto.label];
    switch_val -> [Jmp#switch_val.fail_label|
		   lists:map(fun (C) -> element(2,C) end,
			     Jmp#switch_val.cases)];
    switch_tuple_arity -> [Jmp#switch_tuple_arity.fail_label|
			   lists:map(fun (C) -> element(2,C) end,
				     Jmp#switch_tuple_arity.cases)];
    type -> [Jmp#type.true_label, Jmp#type.false_label];
    call -> [Jmp#call.continuation_label| 
	     case Jmp#call.fail_label of [] -> []; L -> [L] end];
    pushcatch -> [Jmp#pushcatch.successor, Jmp#pushcatch.label];
    _ -> []
  end.

%%
%% @doc Redirects jumps from label Old to label New. If the
%%      instruction does not jump to Old, it remains unchanged.
%%

redirect_jmp(Jmp, ToOld, ToNew) ->
  NewIns =
    case type(Jmp) of
      'if' ->
	NewJmp = if ToOld =:= Jmp#'if'.true_label -> Jmp#'if'{true_label=ToNew};
		    true -> Jmp
		 end,
	if ToOld =:= NewJmp#'if'.false_label -> NewJmp#'if'{false_label=ToNew};
	   true -> NewJmp
	end;
      switch_val ->
	NewJmp = if ToOld =:= Jmp#switch_val.fail_label -> 
		     Jmp#switch_val{fail_label=ToNew};
		    true -> Jmp
		 end,
	NewJmp#switch_val{cases = 
			  lists:map(fun (Pair) ->
					case Pair of 
					  ({Val,ToOld}) -> {Val, ToNew};
					  (Unchanged) -> Unchanged
					end
				    end, 
				    NewJmp#switch_val.cases)
			 };
      switch_tuple_arity ->
	NewJmp = if ToOld =:= Jmp#switch_tuple_arity.fail_label -> 
		     Jmp#switch_tuple_arity{fail_label=ToNew};
		    true -> Jmp
		 end,
	NewJmp#switch_tuple_arity{cases = 
				  lists:map(fun (Pair) -> 
						case Pair of
						  ({Val,ToOld}) -> {Val, ToNew};
						  (Unchanged) -> Unchanged
						end
					    end, 
					    NewJmp#switch_tuple_arity.cases)
				 };
      type ->
	NewJmp = if ToOld =:= Jmp#type.true_label -> Jmp#type{true_label=ToNew};
		    true -> Jmp
		 end,
	if ToOld =:= NewJmp#type.false_label -> NewJmp#type{false_label=ToNew};
	   true -> NewJmp
	end;
      goto ->
	if ToOld =:= Jmp#goto.label -> Jmp#goto{label=ToNew};
	   true -> Jmp
	end;
      call -> 
	NewCont = case Jmp#call.continuation_label of
		    ToOld -> ToNew;
		    OldCont -> OldCont
		  end,
	NewFail = case Jmp#call.fail_label of
		    ToOld -> ToNew;
		    OldFail -> OldFail
		  end,
	Jmp#call{continuation_label = NewCont, 
		 fail_label = NewFail};
      pushcatch ->
	NewLabl =	case pushcatch_label(Jmp) of
			  ToOld ->  ToNew;
			  OldLab -> OldLab
			end,
	NewSucc =	case pushcatch_successor(Jmp) of
			  ToOld ->  ToNew;
			  OldSucc -> OldSucc
			end,
	Jmp#pushcatch{label = NewLabl,successor=NewSucc};
      _ ->  Jmp
    end,
  simplify_branch(NewIns).

%%
%% @doc Turns a branch into a goto if it has only one successor and it
%%      is safe to do so.
%%

simplify_branch(I) ->
  case ordsets:from_list(successors(I)) of
    [Label] ->
      Goto = mk_goto(Label),
      case type(I) of
	type -> Goto;
	'if' -> Goto;
	switch_tuple_arity -> Goto;
	switch_val -> Goto;
	_ -> I
      end;
    _ -> I
  end.

%%
%% Is this an unconditional jump (causes a basic block not to have a 
%% fallthrough successor).
%%

%% is_uncond(I) ->
%%   case type(I) of
%%     goto -> true;
%%     fail -> true;
%%     enter -> true;
%%     return -> true;
%%     call -> 
%%       case call_fail(I) of
%% 	[] -> 
%% 	  case call_continuation(I) of
%% 	    [] -> false;
%% 	    _ -> true
%% 	  end;
%% 	_ -> true
%%       end;
%%     _ -> false
%%   end.

%% @spec is_branch(icode_instruction()) -> bool()
%%
%% @doc Succeeds if the Icode instruction is a branch. I.e. a
%%      (possibly conditional) discontinuation of linear control flow.
%% @end

is_branch(Instr) ->
  case type(Instr) of
    'if' -> true;
    switch_val -> true;
    switch_tuple_arity -> true;
    type -> true;
    goto -> true;
    fail -> true;
    call -> 
      case call_fail(Instr) of
	[] -> 
	  case call_continuation(Instr) of
	    [] -> false;
	    _ -> true
	  end;
	_ -> true
      end;
    enter -> true;
    return -> true;
    pushcatch -> true;
    _ -> false
  end.

%%
%% @doc Makes a new variable.
%%

mk_new_var() ->
  mk_var(hipe_gensym:get_next_var(icode)).

%%
%% @doc Makes a new fp variable.
%%

mk_new_fvar() ->
  mk_fvar(hipe_gensym:get_next_var(icode)).

%%
%% @doc Makes a new register.
%%

mk_new_reg() ->
  mk_reg(hipe_gensym:get_next_var(icode)).

%%
%% @doc Makes a new label.
%%

mk_new_label() ->
  mk_label(hipe_gensym:get_next_label(icode)).

%%
%% @doc Makes a bunch of move operations.
%%

mk_moves([], []) ->
  [];
mk_moves([X|Xs], [Y|Ys]) ->
  [mk_move(X, Y) | mk_moves(Xs, Ys)].

%%
%% Makes a series of element operations.
%%

%% mk_elements(_, []) -> 
%%   [];
%% mk_elements(Tuple, [X|Xs]) ->
%%   [mk_primop([X], {unsafe_element, length(Xs)+1}, [Tuple]) | 
%%    mk_elements(Tuple, Xs)].


%%
%% @doc Converts old Icode to new Icode.
%%  Compliant with calls ending basic blocks.
%%

fixup_fallthroughs(Icode) ->
  Code = icode_code(Icode),
  CompliantCode = fixup_fallthroughs(Code, []),
  icode_code_update(Icode, CompliantCode).

fixup_fallthroughs([I0|Is0], Acc) ->
  {I,Is} = fixup_one_fallthrough(I0, Is0),
  fixup_fallthroughs(Is, [I|Acc]);
fixup_fallthroughs([], Acc) ->
  lists:reverse(Acc).

fixup_one_fallthrough(I0, Is0 = [I1|_]) ->
  case is_label(I1) of
    false ->
      {I0,Is0};
    true ->
      case is_branch(I0) of
	true ->
	  {I0,Is0};
	false -> %% non-branch before label, doubleplusungood
	  %% io:format("~w: fixed unsafe fallthrough <~w,~w>\n", [?MODULE,I0,I1]),
	  {I0,[mk_goto(label_name(I1)) | Is0]}
      end
  end;
fixup_one_fallthrough(I0, []) ->
  {I0,[]}.


%%
%% @doc Removes comments from Icode.
%%

strip_comments(ICode) ->
  icode_code_update(ICode, no_comments(icode_code(ICode))).

no_comments([]) ->
  [];
no_comments([I|Xs]) ->
  case is_comment(I) of 
    true -> no_comments(Xs);
    false -> [I|no_comments(Xs)]
  end.

%%-----------------------------------------------------------------------

%% @spec is_safe(icode_instruction()) -> bool()
%%
%% @doc True if an Icode instruction is safe (can be removed if the
%% result is not used).

is_safe(Instr) ->
  case type(Instr) of
    move -> true;
    fmove -> true;
    phi -> true;
    restore_catch -> true;
    call ->
      case call_fun(Instr) of
	{M,F,A} ->
	  erl_bifs:is_safe(M,F,A);
	Op ->
	  hipe_icode_primops:is_safe(Op)
      end;
    _ -> false
  end.

%% @spec is_leaf_code([icode_instruction()]) -> bool()
%% @doc Returns <code>false</code> if code sequence is not leaf code,
%% otherwise <code>true</code>.

is_leaf_code([]) ->
  true;
is_leaf_code([I|Is]) ->
  case type(I) of
    call ->
      case call_type(I) of
	primop -> 
	  case call_fun(I) of
	    call_fun -> false;		% Calls closure
	    enter_fun -> false;		% Calls closure
	    _ -> is_leaf_code(Is)	% Other primop calls are ok
	  end;
	_ -> 
	  false    % non-primop call
      end;
    enter -> false;
    _ ->
      is_leaf_code(Is)
  end.

%%-----------------------------------------------------------------------

highest_var(Code) ->
  highest_var(Code, 0).

highest_var([I|Is], Max) ->
  Defs = defines(I),
  Uses = uses(I),
  highest_var(Is, new_max(Defs++Uses,Max));
highest_var([], Max) ->
  Max.

new_max([V|Vs], Max) ->
  VName = 
    case is_var(V) of
      true ->
	var_name(V);
      false ->
	case is_fvar(V) of
	  true ->
	    fvar_name(V);
	  _ ->
	    reg_name(V)
	end
    end,
  if VName > Max ->
      new_max(Vs, VName);
     true ->
      new_max(Vs, Max)
  end;
new_max([], Max) ->
  Max.

%%-----------------------------------------------------------------------

highest_label(Code) ->
  highest_label(Code, 0).

highest_label([I|Is], Max) ->
  case is_label(I) of 
    true ->
      L = label_name(I),
      NewMax = if L > Max -> L; true -> Max end,
      highest_label(Is, NewMax);
    false ->
      highest_label(Is, Max)
  end;
highest_label([], Max) ->
  Max.

%%-----------------------------------------------------------------------
